from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import warnings
import copy
import numbers
import re
import numpy as np
from random import randint, randrange, sample
from functools import partial
from collections import defaultdict

from mmgroup.tests.groups.mgroup_n import MGroupNWord, StdMGroupN
from mmgroup.structures.abstract_rep_space import mod_rand_invertible
from mmgroup.structures.abstract_rep_space import mod_rand_unit
from mmgroup.structures.abstract_rep_space import mod_pwr2
from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepVector
from mmgroup.structures.abstract_mm_rep_space import AbstractMmRepSpace
from mmgroup.structures.abstract_mm_rep_space import add_vector
from mmgroup.structures.mm_space_indices import purge_sparse_entry
from mmgroup.structures.parse_atoms import AtomDict
from mmgroup.structures.parse_atoms import eval_atom_expression  
from mmgroup.structures.parse_atoms import ihex 
from mmgroup.tests.spaces.rep_aux import pm_mat_from_function 
from mmgroup.tests.spaces.rep_aux import pm_diag_from_function 


from mmgroup import mat24 as m24
from mmgroup import generators as gen
from mmgroup.generators import gen_xi_w2_gray as w2_gray


######################################################################
# Modelling a vector
######################################################################

class SparseMmVector(AbstractMmRepVector):
    group = StdMGroupN
    __slots__ =  'p', 'data'
    ERR_P = "Illegal modulus %d for MM vector space"

    def __init__(self, p, tag = 0, i0 = None, i1 = None):
        if (p & 1) == 0 or not 2 < p < 256:
            raise ValueError(self.ERR_P % p)
        self.p = p
        self.data = defaultdict(int)
        add_vector(self, tag, i0, i1)

    def set_zero(self, p):
        if (p & 1) == 0 or not 2 < p < 256:
            raise ValueError(self.ERR_P % p)
        self.p = p
        self.data = defaultdict(int)

    def __len__(self):
        return len(self.data) 
  

######################################################################
# Monomial operation of the group
######################################################################

p, x, y, t, l = 0x10, 0x20, 0x30, 0x40, 0x50

gtag_dict = { 'p':p, 'x':x, 'y':y, 't':t, 'l':l}

A, B, C, T, X, Z, Y = 1, 2, 3, 4, 5, 6, 7

######################################################################
# Operation of pi (tag = 'p')

def mul_Xp(tag, d, i, g):
    eps, pi, rep =  g.cocode, g.perm, g.rep   
    d1 = m24.op_ploop_autpl(d, rep)
    i1 = pi[i]
    s = d1 >> 12
    if eps & 0x800:  # if eps is odd:
        s ^= m24.scalar_prod(d, m24.vect_to_cocode(1 << i))
        s ^= m24.pow_ploop(d, 2) >> 12
    return s, tag, d1 & 0x7ff, i1


def mul_Yp(tag, d, i, g):
    eps, pi, rep =  g.cocode, g.perm, g.rep   
    d1 = m24.op_ploop_autpl(d, rep)
    i1 = pi[i]
    s = (d1 >> 12) ^ (d1 >> 11)
    if eps & 0x800:  # if eps is odd:
        tag = Z
        s ^= d1 >> 11
    return s, tag, d1 & 0x7ff, i1

def mul_Zp(tag, d, i, g):
    eps, pi, rep =  g.cocode, g.perm, g.rep   
    d1 = m24.op_ploop_autpl(d, rep)
    i1 = pi[i]
    s = d1 >> 12
    if eps & 0x800:  # if eps is odd:
        tag = Y
        s ^= d1 >> 11
    return s, tag, d1 & 0x7ff, i1



def mul_Tp(tag, octad, sub, g):
    d = m24.octad_to_gcode(octad)
    c = m24.suboctad_to_cocode(sub, d)
    eps, pi, rep =  g.cocode, g.perm, g.rep   
    d1 = m24.op_ploop_autpl(d, rep)
    c1 = m24.op_cocode_perm(c, pi)
    s = d1 >> 12
    s ^= (eps >> 11) & 1 & m24.suboctad_weight(sub) 
    octad1 = m24.gcode_to_octad(d1) 
    sub1 = m24.cocode_to_suboctad(c1, d1)
    return s, tag, octad1, sub1

def mul_Ap(tag, i, j, g):
    pi =  g.perm
    i1, j1 = pi[i], pi[j]
    i1, j1 = max(i1, j1), min(i1, j1)
    return  0, tag, i1, j1 

mul_Bp = mul_Ap

def mul_Cp(tag, i, j, g):
    eps, pi =  g.cocode, g.perm
    i1, j1 = pi[i], pi[j]
    i1, j1 = max(i1, j1), min(i1, j1)
    return  (eps >> 11) & 1, tag, i1, j1 



######################################################################
# Operation of x (tag = 'x')

def mul_Xx(tag, d, i, g):
    e = g.pl
    s = m24.scalar_prod(e, m24.vect_to_cocode(1 << i))
    s ^= m24.ploop_comm(d,e)
    return s, X, d & 0x7ff, i



def mul_Yx(tag, d, i, g):
    e = g.pl
    ed = m24.mul_ploop(m24.pow_ploop(e,3), d)
    s = m24.scalar_prod(e, m24.vect_to_cocode(1 << i))
    s = (ed >> 12) ^  (ed >> 11)  
    return s & 1, Y, ed & 0x7ff, i


def mul_Zx(tag, d, i, g):
    e = g.pl
    ed = m24.mul_ploop(m24.pow_ploop(e,3), d)
    s = m24.scalar_prod(e, m24.vect_to_cocode(1 << i))
    s = (ed >> 12)  
    return s & 1, Z, ed & 0x7ff, i

def mul_Tx(tag, octad, sub, g):
    d = m24.octad_to_gcode(octad)
    c = m24.suboctad_to_cocode(sub, d)
    e = g.pl
    s = m24.ploop_comm(d, e)
    s ^= m24.scalar_prod(e, c)
    return s, tag, octad, sub

def mul_Ax(tag, i, j, g):
    return 0, tag, i, j

def mul_Bx(tag, i, j, g):
    c = m24.vect_to_cocode((1 << i) ^ (1 << j))
    e = g.pl
    s = m24.scalar_prod(e, c)
    return s, tag, i, j

mul_Cx = mul_Bx


######################################################################
# Operation of y (tag = 'y')


def mul_Xy(tag, d, i, g):
    e = g.pl
    de = m24.mul_ploop(d, e)
    s =  de >> 12
    return s, X, de & 0x7ff, i


def mul_Yy(tag, d, i, g):
    e = g.pl & 0xfff
    s = m24.scalar_prod(e, m24.vect_to_cocode(1 << i))
    s ^= m24.ploop_comm(d,e) 
    return s, Y, d & 0x7ff, i


def mul_Zy(tag, d, i, g):
    e = g.pl
    de = m24.mul_ploop(d, e)
    s = m24.scalar_prod(e, m24.vect_to_cocode(1 << i))
    s ^= de >> 12
    return s, Z, de & 0x7ff, i


def mul_Ty(tag, octad, sub, g):
    d = m24.octad_to_gcode(octad)
    c = m24.suboctad_to_cocode(sub, d)
    e = g.pl
    c1 = m24.ploop_cap(d, e) ^ c
    sub1 = m24.cocode_to_suboctad(c1, d)
    s = m24.scalar_prod(e, c)
    return s, tag, octad, sub1


def mul_Ay(tag, i, j, g):
    c = m24.vect_to_cocode((1 << i) ^ (1 << j))
    e = g.pl
    s = m24.scalar_prod(e, c)
    if i == j: assert s == 0
    return s, tag, i, j 

def mul_By(tag, i, j, g):
    c = m24.vect_to_cocode((1 << i) ^ (1 << j))
    e = g.pl
    s = m24.scalar_prod(e, c)
    tag = B + s
    return s, tag, i, j

def mul_Cy(tag, i, j, g):
    c = m24.vect_to_cocode((1 << i) + (1 << j))
    e = g.pl
    s = m24.scalar_prod(e, c)
    tag = C - s
    return s, tag, i, j



######################################################################
# Operation of t (tag = 't')

def sign_t_XY(d, i):
    return m24.scalar_prod(d, m24.vect_to_cocode(1 << i))

def sign_t_YZ(d, i):
    return (m24.scalar_prod(d, m24.vect_to_cocode(1 << i))
            ^ (m24.pow_ploop(d, 2)) >> 12)

def sign_t_ZX(d, i):
    return m24.pow_ploop(d, 2) >> 12

def sign_t_0(d,i):
    return 0

XYZ_t_TAGS = [X, Y, Z, X, Y, Z]

XYZ_t_SIGNS = [
    (sign_t_0, sign_t_XY,sign_t_ZX),
    (sign_t_0, sign_t_YZ,sign_t_XY),
    (sign_t_0, sign_t_ZX,sign_t_YZ),
]

def mul_XYZt(index, tag, d, i, g):
    e = g.exp % 3
    return  XYZ_t_SIGNS[index][e](d,i), XYZ_t_TAGS[index + e], d, i

mul_Xt = partial(mul_XYZt, 0)
mul_Yt = partial(mul_XYZt, 1)
mul_Zt = partial(mul_XYZt, 2)



######################################################################
# Monomial operation of l on tags B, C and T
    
# The following dict maps keys (tag, i1, i2) to pairs
# (sign, (tag, i1, i2)), with (-1)**sign the sign of the monomial
# operation. The dictionary comprises all unit vectors for tags
# B, C and the unit vectors with tag T corresponding to even octads.
# These are the firt 375 octads with tag T 
dict_BCT = [None, {}, {}]


table_BCT = [None, (0, gen.make_table(3, 1)),
                 (1024, gen.make_table(3, 2)) ]



def entry_to_atom_box1(x):
    sign = (x >> 15) & 1
    x &= 0x7fff
    if x >= 1536:
        x1 = x - 1536
        return sign, (T, x1 >> 6, x1 & 63)
    elif x >= 768:
        return sign, (C, (x >> 5) - 24, x & 31)
    else:
        return sign, (B, (x >> 5), x & 31)

def entry_to_atom_box2(x):
    sign = (x >> 15) & 1
    return sign, (T, ((x >> 6) & 0x1ff) + 15, x & 63)


def unpack(a):
    return (a[0],) + a[1]

def make_dict_BCT():
    global dict_BCT
    #print("Computing dict...",  end = "")
    ar_T = [None, gen.make_table(1, 1), gen.make_table(1, 2)]
    for i in range(2496):
        t1 =  ar_T[1][i]
        if t1:
            key_ = entry_to_atom_box1(i)[1]
            dict_BCT[1][key_] = unpack(entry_to_atom_box1(t1))
            dict_BCT[2][key_] = unpack(entry_to_atom_box1(ar_T[2][i]))
    #print("length =", len(dict_BCT[1]), 375*64+2*24*24)
    ar_T = [None, gen.make_table(2, 1), gen.make_table(2, 2)]
    for i in range(360 * 64):
        key_ = entry_to_atom_box2(i)[1]
        dict_BCT[1][key_] = unpack(entry_to_atom_box2(ar_T[1][i]))
        dict_BCT[2][key_] = unpack(entry_to_atom_box2(ar_T[2][i]))


make_dict_BCT()


def op_l_BCT(tag, i, j, g):
    e = g.exp % 3
    if e == 0:
        return 0, tag, i, j
    try:
        return dict_BCT[e][tag, i, j]
    except KeyError:
        tag, o, s = tag, i, j
        assert tag == T, tag
        ofs, table = table_BCT[e]
        r = table[((o - 375) << 6) + s]
        sign = (r >> 15) & 1
        o = ofs + ((r >> 5) & 0x3ff)
        return sign, X, o, r & 0x1f


mul_Bl = mul_Cl = mul_Tl = op_l_BCT 


######################################################################
# Monomial operation of l on tag X


tag_ofs_shift = [None, None, None, (T,375,6), (X,0,5), (X,1024,5)]

def make_table_x_entry(exp, d):
     box = d + 4
     table = gen.make_table(box, exp)
     box = gen.gen_xi_op_xi_short(box << 16, exp) >> 16
     tag, ofs, shift  = tag_ofs_shift[box]
     mask = (1 << shift) - 1
     return tag, ofs, shift, mask, table


table_X = [None, [[],[]], [[],[]]]

def make_table_X():
    global table_X
    for exp in (1,2):
        for d in (0,1):
            table_X[exp][d] = make_table_x_entry(exp, d)

make_table_X()
    
    
def mul_Xl(tag, d, i, g):
    e = g.exp % 3
    if e == 0:
        return 0, tag, d, i
    tag, ofs, sh, mask, table = table_X[e][d >> 10]
    r = table[ ((d & 0x3ff) << 5) + i]
    sign = (r >> 15) & 1
    d = (r & 0x7fff) >> sh
    i = r & mask
    return sign, tag, d + ofs, i





######################################################################
# Dictionary of monomial operations

mult_monomial_dict = {}

for vtag0, v in enumerate("ABCTXZY"):
    vtag = vtag0 + 1
    for g in "pxytl":
        gtag = gtag_dict[g]
        try:
            mult_monomial_dict[vtag + gtag] = globals()["mul_" + v + g]
        except:
            pass


  

######################################################################
# Non-monomial operation of the group
######################################################################


def zero_vector(length, *data):
    """Return zero numpy vector of given length and type np.int32"""
    return np.zeros(length, dtype = np.int32)


######################################################################
# Non-onomial operation of t


class NonMonomialOp_t:
    """Helper for class NonMonomialOp for the group element t.

    Implements the non-momomial operation of the group element t**e
    on vectors with tag "ABC" and "T".
 
    General Interface see class MonomialOp_any. See class 
    NonMonomialOp for background.
    """
    __slots__ = "d_ABC", "d_T", "d_D", "exp", "p", "half", "eighth"
    mat3_e1 = np.array([[0, 2,  -2], [1, 1, 1], [1,  -1, -1]],
          dtype = np.int32)
    mat3_e2 = np.array([[0, 2,  2], [1, 1, -1], [-1,  1, -1]],
          dtype = np.int32)
    mat3 = [None, mat3_e1 , mat3_e2]
    diag64 = pm_diag_from_function(m24.suboctad_weight, 64)
    sym64 = pm_mat_from_function(m24.suboctad_scalar_prod, 64)
    mat64_e1 = sym64.dot(diag64)
    mat64_e2 = diag64.dot(sym64)
    # TODO: explain why mat64_e1, mat64_e2 are not assigned vice versa!!!
    mat64 = [None, mat64_e1, mat64_e2] 
    for m in mat64[1:]:
        assert (m.dot(m).dot(m) == 512 * np.identity(64)).all()
    assert (mat64_e1.dot(mat64_e2) == 64* np.identity(64)).all()
 

    def __init__(self, p, exp):
        self.d_ABC = defaultdict(partial(zero_vector,3))
        self.d_T = defaultdict(partial(zero_vector,64))
        self.d_D = defaultdict(int)
        self.exp = exp % 3
        self.p = p
        self.half = mod_pwr2(-1, p)
        self.eighth= mod_pwr2(-3, p)
        assert exp in (1,2)

    def load(self, tag, i, j, coeff): 
        coeff %= self.p
        if tag == T:
            self.d_T[i][j] = coeff 
        elif tag in (A, B, C):
            if i == j:
                assert tag == A
                self.d_D[i] = coeff
            else:
                assert i > j, (tag, (i,j))
                self.d_ABC[(i,j)][tag-1] = coeff
        else:
            raise ValueError("Bad tag %s" % tag)

    def op_ABC(self):
        m = self.mat3[self.exp]
        for (i,j), v in self.d_ABC.items():
            assert i > j, ("* ", v, (i,j))
            v = np.dot(v, m) * self.half % self.p
            if v[0]: yield A, i, j, int(v[0])      
            if v[1]: yield B, i, j, int(v[1])      
            if v[2]: yield C, i, j, int(v[2])      
               
    def op_T(self):
        m = self.mat64[self.exp]
        for oct, v in self.d_T.items():
            v = np.dot(v, m) * self.eighth % self.p
            for i in range(64):
                if v[i]: yield T, oct, i, int(v[i]) 

    def process(self):
        for i, value in self.d_D.items():
            if value:
                yield A, i, i, value
        yield from self.op_ABC()
        yield from self.op_T()


######################################################################
# Non-monomial operation of l



class NonMonomialOp_l:
    """Helper for class Space4096x for the group element l.

    Implements the non-momomial operation of the group element l**e
    on vectors with tag  "A", "Y", "Z".
 
    General Interface see class auto_group_space.UnprocessedVector. See 
    class Space4096x and its base class mvector.AutoGroup_RepSpace
    for background.
    """
    __slots__ = "d_YZ", "d_A", "exp", "p", "quater", "eighth"
    # Prepare 16 x 16 matrix for tag A
    # Put L4A = 2 * xi_24a with xi_24a as in [Seys19], section 9
    L4A = pm_mat_from_function(lambda i, j: i == j, 4)
    # Put L4B = xi_24b with xi_24b as in [Seys19], section 9
    L4B = pm_diag_from_function(lambda i: i > 0, 4)
    L4_POWERS = [L4A @ L4B, L4B @ L4A]
    mat16 = [None] + [np.kron(m4, m4) for m4 in L4_POWERS]

    # Prepare 64 x 64 matrix for tags Y and Z
    MSYM16 = pm_mat_from_function(lambda i, j: w2_gray(i ^ j), 16)
    MDIAG16 = -pm_diag_from_function(lambda i: w2_gray(i), 16)
    L16_POWERS = [MDIAG16 @ MSYM16, MSYM16 @ MDIAG16]
    powers_mat64 = zip(L16_POWERS, L4_POWERS)
    mat64 = [None] + [np.kron(i,j) for i,j in powers_mat64]
    cycles3 = [None, [2, 0, 1, 3], [1, 2, 0, 3]] 

    def __init__(self, p, exp):
        self.d_YZ = defaultdict(partial(zero_vector,64))
        self.d_A = defaultdict(partial(zero_vector,16))
        self.exp = exp % 3
        self.p = p
        self.quater = mod_pwr2(-2, p)
        self.eighth= mod_pwr2(-3, p)
        assert exp in (1,2)
        

    def load(self, tag, i, j, coeff): 
        coeff %= self.p
        if tag in (Y, Z):
            i += (tag == Y) << 11
            ih, il = i & 0xff0, i & 0xf
            jh, jl = j & 0x1c, j & 3
            self.d_YZ[(ih, jh)][4 * il + jl] = coeff 
            return
        if tag == A:
            ih, il = i & 0x1c, i & 3
            jh, jl = j & 0x1c, j & 3
            self.d_A[(ih,jh)][4 * il + jl] = coeff
            if ih == jh:
                self.d_A[(ih,jh)][4 * jl + il] = coeff
            return
        raise ValueError("Bad tag %s for operation of xi" % tag)

    def op_YZ(self):
        m = self.mat64[self.exp]
        cyc3 = self.cycles3[self.exp]
        for (ih, jh), v in self.d_YZ.items():
            v = np.dot(v, m) * self.eighth % self.p
            k_hi = cyc3[ih >> 10]
            ih = ((k_hi) << 10) + (ih & 0x3ff)
            tag = Z + (ih >> 11)
            ih &= 0x7ff
            for il in range(16):
                for jl in range(4):
                    x = int(v[4 * il + jl]) 
                    if x: yield  tag, ih + il, jh + jl, x 

    def op_A(self):
        m = self.mat16[self.exp]
        for (ih, jh), v in self.d_A.items():
            v = np.dot(v, m) * self.quater % self.p
            assert ih >= jh
            if ih > jh:
               for il in range(4):
                   for jl in range(4):
                       x = int(v[4 * il + jl]) 
                       if x: yield A, ih + il, jh + jl, x 
            else:
               for il in range(4):
                   for jl in range(il+1):
                       x = int(v[4 * il + jl]) 
                       if x: yield A, ih + il, jh + jl, x

    def process(self):
        yield from self.op_YZ()
        yield from self.op_A()

######################################################################
# Dictionary of non-monomial operations


non_monomial_dict = {
   "t": NonMonomialOp_t,
   "l": NonMonomialOp_l,
}


######################################################################
# The default instance of the monster group
######################################################################

default_monster_group = StdMGroupN

try:
    from mmgroup.mm_space import standard_mm_group
except ImportError:
    default_monster_group.target_group = None
    err = "C version of monster group support not implemented"
    warnings.warn(err, UserWarning)
else:
    default_monster_group.target_group = standard_mm_group 
    #default_monster_group.set_preimage(standard_mm_group, tuple)

######################################################################
# class SparseMmSpace
######################################################################




class SparseMmSpace(AbstractMmRepSpace):
    """Models the sparse representation 198884x of the monster group. 

    YET TO BE DOCUMENTED !!!

    """
    group = StdMGroupN
    vector_type = SparseMmVector
    space_name = "MVSp"

    def __init__(self):
        """Create a 196884-dimensional representation of the monster

        All calculations are done modulo the odd number p
        """
        pass

    #######################################################################
    # Creating vectors 
    #######################################################################


    def zero(self, p):
        return self.vector_type(p, 0)
        

    def copy_vector(self, v1):
        """Return deep copy of group element v1"""
        v2 = self.vector_type(v1.p, 0)
        v2.data.update(v1.data)
        return v2


    def __call__(self, p, tag = 0, i0 = None, i1 = None):
        return self.vector_type(p, tag, i0, i1)

    #######################################################################
    # getitem and setitem
    #######################################################################



    def getitems_sparse(self, v1, a_sparse):
        """Get items from vector v1

        Here we assert that v1 is a vector of this vector space and
        that 'a_sparse' is a one-dimensional numpy array of type
        np.uint32, containing the coordinates to be read from v1.
 
        The function must add the corresponding coordinate to each
        entry of the array 'a_sparse' to v1. All coordinates must be
        nonnegative and < 256. 

        A zero entry in the array 'a_sparse' should be ignored.
        """
        assert v1.space == self
        for i, sp in enumerate(a_sparse):
            index = purge_sparse_entry(sp) 
            if index & 0xffffff00:
                a_sparse[i] = sp  ^ (v1.data[index] & 0xff)
        return  a_sparse 

    def setitems_sparse(self, v1, a_sparse):
        """Set items from vector v1

        Here we assert that v1 is a vector of this vector space and
        that 'a_sparse' is a one-dimensional numpy array of type
        np.uint32, containing the coordinates and values to be 
        written to v1.
 
        The function must set the corresponding coordinates in v1. 
        All coordinates must be nonnegative and < 256. 

        A zero entry in the array 'a_sparse' should be ignored.
        """
        assert v1.space == self
        for sp in a_sparse:
            sp = purge_sparse_entry(sp) 
            index = int(sp & 0xffffff00)
            if index:                 
                v1.data[index] = int(sp & 0xff)
        self.reduce(v1)
        return v1

    def additems_sparse(self, v1, a_sparse):
        """Add a vector in sparse representation to vector v.

        This method takes a numpy array 'a_indices' of integers of dtype 
        numpy.uint32 containing the description of a vector v2 in sparse 
        representation. It computes 

             v  =  v + v2 .

        Here vector v is a standard vector in this space.
        """
        assert v1.space == self
        for sp in a_sparse:
            sp = purge_sparse_entry(sp) 
            index = sp & 0xffffff00
            if index:                 
                v1.data[index] += int(sp & 0xff)
        self.reduce(v1)
        return v1



    #######################################################################
    #  vector addition and scalar multiplication 
    #######################################################################
 
    def reduce(self, v1):
        to_delete = []
        data = v1.data
        for sp, value in data.items():
            value %= v1.p
            data[sp] = value
            if value == 0:
                to_delete.append(sp)
        for sp in to_delete:
            del data[sp]
        return v1

    def iadd(self, v1, v2):
        for sp, value in v2.data.items():
            v1.data[sp] += value
        return self.reduce(v1)
 
    def imul_scalar(self, v1, a):
        a = int(a % v1.p)
        for sp, value in v1.data.items():
            v1.data[sp] = value * a
        return self.reduce(v1)

    #######################################################################
    # Obtaining a vector in sparse format 
    #######################################################################
               

    def iter_as_sparse(self, v1):
        for key_, value in v1.data.items():
            value %= v1.p
            if value:
                yield key_ + value

    def as_sparse(self, v1):
        return np.fromiter(self.iter_as_sparse(v1), dtype = np.uint32)
           
    #######################################################################
    # Multiplying with a group element
    #######################################################################

    def imul_group_monomial(self, v, g_list, w):
        p = v.p
        g_tag_list = [(g, gtag_dict[g.tag]) for g in g_list]
        wd = w.data
        wd.clear()
        for vu, v_coord in v.data.items():
            sign = 0
            vtag, i, j = vu >> 25, (vu >> 14) & 0x7ff, (vu >> 8) & 0x3f
            for g, g_tag in g_tag_list:
                f = mult_monomial_dict[vtag + g_tag] 
                s, vtag, i, j = f(vtag, i, j, g)
                sign ^=  s
            wu = int((vtag << 25) + (i << 14) + (j << 8))
            wd[wu] = p - v_coord if (sign & 1) else v_coord
        
            

    def imul_group_atom(self, v, g, w):
        p = v.p
        w.data.clear()
        gtag = gtag_dict[g.tag]
        if g.tag in "tl":
            e = int(g.exp % 3)
            if e == 0:
                w.data.update(v.data)
                return
            non_monomial = non_monomial_dict[g.tag](p, e)
        for vu, v_coord in v.data.items():
            vtag, i, j = vu >> 25, (vu >> 14) & 0x7ff, (vu >> 8) & 0x3f
            try:
                f = mult_monomial_dict[vtag + gtag]
            except KeyError:
                non_monomial.load(vtag, i, j, v_coord)
            else:
                sign, wtag, wi, wj = f(vtag, i, j, g)
                wu = int((wtag << 25) + (wi << 14) + (wj << 8))
                #assert 0 <= v_coord <= p                
                w.data[wu] = p - v_coord if (sign & 1) else v_coord
        if g.tag in "tl":
            for wtag, wi, wj, w_coord in non_monomial.process():
                wu = int((wtag << 25) + (wi << 14) + (wj << 8))
                w.data[wu] = w_coord
 
       
    def imul_group_word(self, v1, g):
        """Return product v1*g of vector v1 and group word g.

        v1 may be destroyed.

        This method is called for elements v1 of the space
        'self' and for elements g of the group 'self.group' only.
        """
        g_mon = []
        v, w = v1, self.zero(v1.p)
        p = v1.p
        data = v.data
        for i, coord in data.items():
            data[i] = coord % p
        for atom in v1.group(g).iter_generators():
            if not atom.tag in "tl":
                g_mon.append(atom)
            else:
                if len(g_mon):
                    self.imul_group_monomial(v, g_mon, w)
                else:
                    v, w = w, v
                g_mon = []
                self.imul_group_atom(w, atom, v)
        if len(g_mon):
            self.imul_group_monomial(v, g_mon, w)
            v, w = w, v
        return v

    #######################################################################
    # Testing equality of vectors
    #######################################################################


    def equal_vectors(self, v1, v2):
        """Return True iff vectors v1 and v2 are equal 

        This method is called for elements v1 and v2 of the space
        'self' only.
        """
        v1 = v1.reduce()
        v2 = v2.reduce()
        return v1.p == v2.p and v1.data == v2.data
        
        
StdSparseMmSpace = SparseMmSpace()
SparseMmVector.space = StdSparseMmSpace

def SparseMmV(p):
    return partial(SparseMmVector, p)





# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals




include "../pxd_files/mm_basics.pxi"


from mmgroup.generators import  rand_get_seed


######################################################################
### Python functions added for convenience
######################################################################




import numpy


PROTECT_OVERFLOW =  (17 << INT_BITS) // 19

def mmv_array(*args):
    """Create array with data type suitable for a Monster rep vector 

    This is a numpy array of shape given by the positional arguments
    and of type suitable for the monster group.

    For internal purposes only.
    """
    return numpy.zeros(args, dtype = numpy.uint64)

def mm_vector(p, *args):
    """Create zero vector or array suitable for a Monster rep vector 

    Such vectors are implemented as numpy arrays
    E.g.
    mm_vector(3)   # Create a monster rep vector in characteristic p=3
    mm_vector(3,5) # Create an array of 5 such vectors

    The most convenient way to use this function is e.g.:

    from mm_op import MM
    mm3 = MM[3]        # Class containing monster functions for p=3
    v = mm3.vector()   # Create one monster rep vector v for p=3
    mm3.random_mmv(v)  # randomize the vector v
    w = mm3.vector(5)  # Create array of 5 such monster rep vectors 
    w[3] = v           # standard numpy operation
    mm3.get_mmv(v,3)   # get component 3 of v
    """
    assert (p & (p + 1) == 0) and 3 <= p < 256
    vsize = mm_aux_mmv_size(p)
    result = numpy.zeros(args + (vsize + 1,), dtype = numpy.uint64)
    copy = result.reshape(-1, vsize + 1)
    for i in range(copy.shape[0]):
         copy[i, vsize] = PROTECT_OVERFLOW
    return result







def tag_dict_from_compressed(uint32_t p, d, a):
    cdef uint32_t[::1] pa = a
    cdef uint32_t len_a = len(a)
    cdef uint32_t tag, i, i0, i1, x
    for i in range(len_a):
        x = pa[i]
        tag = (x >> 25) - 1
        i0 = (x >> 14) & 0x7ff
        i1 = (x >> 8) & 0x3f
        x &= p
        try:
            d[("ABCTXZY"[tag], i0, i1)] = x
        except IndexError:
            raise IndexError("Bad compressed entry %s" % hex(pa[i]))        
        

def tag_dict_to_compressed(d, a):
    cdef uint32_t[::1] pa = a
    cdef uint32_t len_a = len(a)
    cdef uint32_t i = 0
    cdef uint32_t tag, i0, i1, x
    for  (t, i0, i1), x in d.items():
        tag = "0ABCTXZY".index(t)
        pa[i] = (tag << 25) + (i0 << 14) + (i1 << 8) + x
        i = i + 1
        if i > len_a:
            raise IndexError("Compressed array too short")
    return i

    
    








from __future__ import absolute_import, division, print_function

import os
import sys

import setuptools
from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Distutils import build_ext
from Cython.Build import cythonize


#from distutils.core import setup
#from distutils.extension import Extension
#from Cython.Build import cythonize

FILE_DIR = os.path.dirname(os.path.realpath(__file__))
#print("FILE_DIR ==", FILE_DIR)
TEST_DIR = os.path.dirname(FILE_DIR)
#print("TEST_DIR ==", TEST_DIR)
PACKAGE_DIR = os.path.dirname(TEST_DIR)
#print("PACKAGE_DIR ==", PACKAGE_DIR)
SRC_DIR = os.path.dirname(PACKAGE_DIR)


sys.path.append(SRC_DIR)   # for all the stuff in the 'mmgroup' package
from mmgroup.tests import test_hadamard
from mmgroup.tests.tempdir import temp_dir_name
assert sys.path.pop() == SRC_DIR

C_subdir = temp_dir_name(test_hadamard)


ext_modules=[
    Extension("test_hadamard",
        sources=[
           "test_hadamard.pyx",
           "test_hadamard_op.c",
        ],
        #libraries=["m"], # Unix-like specific
        include_dirs = [ "." ],
    )
]

setup(
  name = "test_hadamard",
  ext_modules =  cythonize(ext_modules)
)




r"""We deal with the Parker loop, which is a non-associative Moufang loop.

The Parker loop is written multiplicatively its elements
can be represented as pairs in
the set :math:`\mathcal{P} = \mathcal{C} \times \mathbb{F}_2`, see
e.g. :cite:`Asc86`, :cite:`Con85`.
We also write ``1`` for the neutral element 
:math:`(0,0) \in  \mathcal{C} \times \mathbb{F}_2 = \mathcal{P}` of
the Parker loop and ``-1`` for its negative :math:`(0,1)`. 
Let :math:`\Omega` = :math:`(0,\tilde{\Omega})`, where 
:math:`\tilde{\Omega}` is the Golay code word ``1,...,1``. Then the 
center :math:`Z(\mathcal{P})` of :math:`\mathcal{P}` is 
:math:`\{1, -1, \Omega, -\Omega\}`.  An element :math:`(g,s)`,
:math:`g \in \mathcal{C}` is called *positive* if :math:`s = 0` 
and *negative* if :math:`s = 1`. 

Multiplication of two Parker loop elements 
:math:`(g_1,s_1), (g_2,s_2)` is given by a cocycle
:math:`\theta: \mathcal{C} \times \mathcal{C} \rightarrow \mathbb{F}_2`
as follows:

 :math:`(g_1,s_1) \cdot (g_2,s_2) = (g_1 + g_2, s_1 + s_2 + \theta(g_1, g_2))`.

There are several possibilities for the cocycle :math:`\theta`. We
choose a cocycle :math:`\theta` satisfying the requirements given 
by :cite:`Seysen20`, see :ref:`basis-golay-label` for details.


We represent elements of :math:`\mathcal{P}` as instances of 
class |PLoop|. For Golay code word  ``g`` given as an instance of 
class |GCode|, the value ``PLoop(g)``  is the positive element 
``(g,0)`` of the Parker loop, and  ``-PLoop(g)`` is the corresponsing 
negative element ``(g,1)``. The constant  ``PLoopOne`` is equal
to the neutral element ``(0,0)`` and the constant ``PLoopOmega``
is equal to the central element ``~PLoopOne``. Here the ``~``
means bitwise complement of the corresponding Golay code element
without changing the sign of a Parker loop element. Thus 
``PLoopOmega`` is the positive Parker loop element corresponding
to the Golay code word containing ``24`` one bits. So the center 
:math:`Z(\mathcal{P})` of :math:`\mathcal{P}` is:

  ``[PLoopOne, -PLoopOne, PLoopOmega, -PLoopOmega]`` .

Let ``g1, g2`` be instances of 
class  |GCode|. For positive elements of the Parker loop the 
multiplication formula given above can be coded as follows: 

  ``PLoop(g1) * PLoop(g2) == (-1)**g1.theta(g2) * PLoop(g1 + g2)`` .

The elements of the Parker loop are numbered from ``0`` to 
``0x1fff``. Here the numbers of the  posivtive Parker loop
elements correspond to the numbers of the Golay code words.
The numbering system for the Parker loop is also used for 
indexing certain elements of the monster group :math:`\mathbb{M}`.

A transversal of :math:`Z(\mathcal{P})` in :math:`\mathcal{P}` is 
used for indexing certain basis vectors of the representation 
:math:`\rho` of :math:`\mathbb{M}`. Property ``ord`` of an 
instance ``a`` of class |Ploop| returns the number of that element 
of the Parker loop. 
For indexing vectors in  :math:`\rho` it is important to 
know that ``(-a).ord =  a.ord ^ 0x1000`` and 
``(~a).ord = a.ord ^ 0x800``. Thus the Parker loop elements
``a`` with ``0 <= a.ord < 0x800`` are a transversal of 
:math:`Z(\mathcal{P})` in :math:`\mathcal{P}`.
"""




from functools import reduce
from operator import __xor__
from numbers import Integral, Number
from random import randint

try:
    # Try importing the fast C function
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24

from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.parity import Parity
from mmgroup.structures.parse_atoms import ihex


from mmgroup.structures.gcode import GCode, GcVector
from mmgroup.structures.cocode import Cocode


ERR_RAND = "Illegal string for constricting type %s element" 

ERR_DIV4 = "%s object may be divided by 2 or 4 only"
ERR_DIV2 = "%s object may be divided by 2 only"


#######################################################################
# Import derived classed
#######################################################################

import_pending = True

def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending
    global AutPL, AutPlGroup, XLeech2
    from mmgroup.structures.autpl import AutPL, AutPlGroup
    from mmgroup.structures.xleech2 import XLeech2
    import_pending = False



#######################################################################
# Auxiliary functions
#######################################################################



def as_vector24(data):
    """Convert list of bit positions to vector in ``GF(2)**24``"""
    try:
        vector = reduce(__xor__, (1 << x for x in data), 0)
    except:
        err = "List of integers >= 0 required for a 24-bit vector"
        raise TypeError(err)
    if vector >= 0x1000000:
        raise ValueError("List entry too large for a 24-bit vector")
    return vector


def str_vector24(v):
    """Return the 24-bit vector ``v`` as a string of binary digits

    Here ``v`` must be an integer with the coefficient of the
    ``i``-th bit coded in the bit with valence ``2**i``.
    Coefficients  ``i >= 24`` are ignored.
    """
    l = ["01"[(v >> i) & 1] for i in range(24)]
    l4 = ("".join(l[i:i+4]) for i in range(0, 24, 4))
    return " ".join(l4)

def str_basis(text, letter, basis):
   s = text + "\n"
   for i, v in enumerate(basis):
       s += "  %3s: %s\n" % (letter + str(i), str_vector24(v)) 
   return s






#######################################################################
# Subclasss PLoop of class GCode
#######################################################################





class PLoop(GCode):
    """This class models an element of the Parker Loop.

    The Parker loop is a non-associative loop with  :math:`2^{1+12}` 
    elements. Each element can be interpreted as a signed Golay
    code word, see class |GCode|. Accordingly, class |PLoop| is
    a subclass of class |GCode|.

    The loop operation is written
    multiplicatively. The neutral element of the loop (corresponding
    to the positive zero word of the Golay code) is ``PLoopOne``, and 
    its negative is ``-PLoopOne``.  ``PLoopOmega`` is the (positive) 
    element corresponding  to the all-one vector of the Golay code. 
        
    The :math:`2^{1+12}` Parker loop elements are numbered from
    ``0`` to ``0x1fff``. Elements ``0`` to ``0xfff`` are are
    positive and corresponding to the Golay code words with the 
    same number. The element with number  ``0x1000 ^ i`` is the
    negative of the element with number  ``i``.

    :param value:

      This is the value of the Parker loop element to be returned. 

    :type value: see table below for legal types

    :return: A Parker loop element
    :rtype:  an instance of class |PLoop|

    :raise:
        * TypeError if ``type(value)`` is not in the table given above.
        * ValueError if ``value`` cannot be converted to an
          instance of class  |PLoop|.


    Depending on its type parameter **value** is  interpreted as follows:

    .. table:: Legal types for constructor of class ``PLoop``
      :widths: 20 80

      ===================== ================================================
      type                  Evaluates to
      ===================== ================================================
      ``int``               Here the code word with number ``value`` is
                            returned.  ``0 <= value < 0x2000`` must hold.
                            We have ``PLoop(0) == PLoopOne``,
                            ``PLoop(0x1000) == -PLoopOne``, and
                            ``PLoop(0x800) == ~PLoopOne == PLoopOmega``.
  
      ``list`` of ``int``   Such a list is converted to a Golay code word,
                            see class |GCode|, and the corresponding 
                            (positive) Parker loop element is returned.

      class |GCode|         The corresponding 
                            (positive) Parker loop element is returned. 

      class |PLoop|         A deep copy of parameter ``value`` is returned.

      class |GcVector|      This is converted to a Golay code word,
                            see class |GCode|, and the corresponding 
                            (positive) Parker loop element is returned.

      class |XLeech2|       The Parker loop part of the |XLeech2| object  
                            is  returned. 

      ``str``               Create random element depending on the string
                             | ``'r'``: Create arbitrary Parker loop element

      ===================== ================================================



    **Standard operations**

    Let ``a`` be a Parker loop element. 
    The multiplication operator ``*`` implements the non-associative
    loop operation. Division by a Parker loop element means multiplication
    by its inverse, and exponentiation means repeated multiplication,
    with ``a**(-1)`` the loop inverse of  ``a``,  as usual. 

    The   ``~``  operator maps the Parker loop element  ``a`` to 
    ``a * PLoopOmega``, leaving the sign of ``a`` unchanged.

    Multiplication with the integer ``1`` or ``-1`` means the 
    multiplication with the neutral element ``PLoopOne`` or 
    with its negative ``-PLoopOne``, respectively.
    
    If any of the oprations  ``+``, ``-`` or ``&`` is applied to a Parker
    loop element, that element is converted to a Golay code word of 
    type |GCode|, ignoring the sign. This conversion also takes place if
    a Parker loop element is multiplied or divided by an integer
    different from ``1`` or ``-1``.
    

    **Standard functions**

    ``len(a)`` is a shorthand for ``len(GCode(a))``.
    Here function ``GCode(a)`` converts  ``a`` to
    a Golay code word of type |GCode|. 

    
    ``abs(a)`` returns the element in the set ``{a, -a}`` which is
    positive.
    """
    __slots__ = "value", "bit_list_"
    parity_class = GCode

    def __init__(self, value = 0):
        if isinstance(value, Integral):
            self.value = value & 0x1fff
            return
        if import_pending:
            complete_import()
        if isinstance(value, PLoop):
            self.value = value.value & 0x1fff
        elif isinstance(value, GCode):
            self.value = value.value & 0xfff
        elif isinstance(value, str):
            self.value = randint(0, 0x1fff)
        elif isinstance(value, GcVector):
            vector = value.value
            vector ^= mat24.syndrome(vector) 
            self.value = mat24.vect_to_gcode(vector)
        elif isinstance(value, XLeech2):
            self.value = (value.value >> 12) & 0x1fff
        else:
            vector = as_vector24(value)
            vector ^= mat24.syndrome(vector) 
            self.value = mat24.vect_to_gcode(vector)

    def __mul__(self, other):
        if isinstance(other, PLoop):
            return PLoop(mat24.mul_ploop(self.value, other.value))
        elif isinstance(other, Integral):
            if abs(other) == 1:
                return PLoop(self.value ^ ((other & 2) << 11))
            return GCode(self.value & -(other & 1))
        elif isinstance(other, AutPL):
            return PLoop(mat24.op_ploop_autpl(self.value, other.rep))
        else:           
            return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, Integral):
            if abs(other) == 1:
                return PLoop(self.value ^ ((other & 2) << 11))
            return GCode(self.value & -(other & 1))
        else:           
            return NotImplemented
                              
    def __truediv__(self, other):
        if isinstance(other, PLoop):
            return PLoop(mat24.mul_ploop(self.value, 
                mat24.pow_ploop(other.value, 3)))
        elif isinstance(other, Integral):
            if abs(other) == 1:
                return PLoop(self.value ^ ((other & 2) << 11))
            elif abs(other) == 2:
                return Parity(0)
            elif abs(other) == 4:
                return Parity(mat24.gcode_weight(self.value) & 1)
            else:
                raise TypeError(ERR_DIV4 % type(self))
        else:           
            return NotImplemented
            
    def __rtruediv__(self, other):
        if isinstance(other, Integral):
            if abs(other) == 1:
                inv = mat24.pow_ploop(self.value, 3)
                return PLoop(inv ^ ((other & 2) << 11))
            else:
                raise TypeError("Can only divide unit by Parker loop element")
        else:           
            return NotImplemented
        
        
    def __pow__(self, other):
        if isinstance(other, Integral):            
            return PLoop(mat24.pow_ploop(self.value, other & 3))
        elif isinstance(other, PLoop):
            comm = mat24.ploop_comm(self.value, other.value) << 12
            return PLoop(self.value ^ comm)
        elif isinstance(other, Parity):
            if mat24.gcode_weight(self.value) & 1:
                raise ValueError("Parker Loop element has order > 2")
            return PLoop(mat24.pow_ploop(self.value, other.value))
        else:
            return NotImplemented

    def __eq__(self, other):
        return (isinstance(other, PLoop)  
            and (self.value ^ other.value) & 0x1fff == 0)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pos__(self):
        return self
        
    def __neg__(self):
        return  PLoop(self.value ^ 0x1000)

    def __invert__(self):
        return  PLoop(self.value ^ 0x800)

    @property
    def ord(self):
        """Return the number of the Parker loop element.

        We have ``0 <= i < 0x2000`` for the returned number ``i``.
        """
        return self.value & 0x1fff


    @property
    def sign(self):
        """Return the sign of the Parker loop element.

        This is ``1`` for a positive and ``-1`` for a negative element.
        """
        return 1 - ((self.value >> 11) & 2)


    def split(self):
        """Split sign and Omega from Parker loop element ``a``. 

        :return:  a triple  ``(es, eo, v)`` with 
                  ``a = (-1)**e1 * PLoopOmega**eo * v`` .

        Here  ``v`` is the unique (positive) element of the Parker  
        loop of type |PLoop| with ``0 <= v.ord < 0x800`` satisfying
        that equation. ``es`` and ``eo`` are ``0`` or ``1``.
        """
        v = self.value
        return (v >> 12) & 1, (v >> 11) & 1, PLoop(v & 0x7ff)
 
    def split_octad(self):
        """Split Parker loop element ``a`` into central element and octad

        :return: a triple ``(es, eo, o)`` with  
                 ``a = (-1)**e1 * PLoopOmega**eo * o``.

        Here ``o`` is either the neutral Parker loop element
        ``PLoopOne`` or Parker loop element corresponding to a
        positive octad.  ``es`` and ``eo`` are ``0`` or ``1``.
        An *octad* is a Golay code word 
        (of type |GCode|)  of weight ``8``.

        :raise:
          * Raise ValueError if this is not possible.
        """
        v = self.value
        e1, eo, w = (v >> 12) & 1, 0, mat24.gcode_weight(v)
        if mat24.gcode_weight(v) > 3:
            v, eo, w = v ^ 0x800, 1, 6 - w
        if w <= 2:
            return e1, eo, PLoop(v & 0xfff)
        raise ValueError("Cannot convert Golay code word to octad")
     
    def str(self):
        return "<PLoop_%s>" % ihex(self.value & 0x1fff, 3)
    __repr__  = str


def PLoopZ(e1 = 0, eo = 0):
    """Return a specific central element of the Parker loop. 

    :param e1: Sign, the sign of the element will be ``(-1)**e1``
    :type e1: int
    :param eo: Exponent for ``PLoopOmega``, default is  ``0``
    :type eo: int
    :return:   The central element ``(-1)**e1  * PLoopOmega**eo``
               of the Parker loop.
    :return type:  an instance of class |PLoop|.

    Here ``PLoopOmega`` is the positive element of the Parker loop
    corresponding to the Golay code word ``1,...,1``.

    ``e1`` and ``eo`` may also be parity objects of type |Parity|.
    """
    e1 = e1.value if isinstance(e1, Parity) else e1
    eo = eo.value if isinstance(eo, Parity) else eo
    return PLoop(((e1 & 1) << 12) | ((eo & 1) << 11))


PLoopOne = PLoop(0)
PLoopOmega = ~PLoopOne    








import sys
import os
import subprocess


if sys.platform.startswith("win32"):
    platform = "WIN32"
else:
    raise NotImplementedError("Platform %s not supported" % sys.platform)


def make_dll_win32(
    name, 
    sources, 
    libraries, 
    include_dirs,
    lib_path,
    dll_path,
    extra_args
    ):
    """Yet to be documented"""
    compile_args = [ "-c", "-Wall", "-DMS_WIN64"] + extra_args 
    for ipath in include_dirs:
        compile_args.append("-I" + os.path.realpath(ipath))
    link_args = ["-shared",  "-Wall", "-DMS_WIN64"] + extra_args 
    objects = []

    # Compile sources and add objects to list 'objects'
    for source in sources:
        args = compile_args[:]
        args.append(os.path.realpath(source))
        args.append("-o")
        obj = os.path.realpath(os.path.splitext(source)[0] + ".o")
        args.append(obj)
        objects.append(obj)
        print("gcc " + " ".join(args))
        subprocess.call("gcc " + " ".join(args)) 

    # link
    largs = link_args[:] + objects 
    for inc_dir in include_dirs:
        # same search path for include files an libraries
        largs.append("-L" + inc_dir)
    for library in libraries: 
        largs.append("-l" + library)
    dll_path =   os.path.join(dll_path, "%s.dll" % name)
    largs +=  ["-o",  dll_path ]
    def_path =   os.path.join(lib_path, "%s.def" % name)
    largs += [ "-Wl,--output-def," + def_path ]
    lib_path =   os.path.join(lib_path, "lib%s.lib" % name)
    largs += [ "-Wl,--out-implib," + lib_path ]
    print("gcc " + " ".join(largs))
    subprocess.call("gcc " + " ".join(largs)) 

    print("removing object files...")
    for obj in objects:
       os.remove(obj)
    print(dll_path + "\nhas been generated.\n")
    


def make_dll(
    name, 
    sources, 
    libraries = [], 
    include_dirs = [],
    lib_path = ".",
    dll_path = ".",
    extra_args = ""
    ):
    if platform == "WIN32":
        return make_dll_win32(name, sources, libraries,  include_dirs,
            lib_path, dll_path, extra_args )


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import pytest

from random import randint, shuffle, sample

from mmgroup import mat24
from mmgroup import GCode, Cocode, PLoop, PLoopZ
from mmgroup import Octad, SubOctad, GcVector 
from mmgroup import Parity
from mmgroup.mat24 import MAT24_ORDER

from mmgroup.tests.spaces.spaces import MMTestSpace 
from mmgroup.mm_space import characteristics
from mmgroup import MM0, MMV

space = MMV
group = MM0

#####################################################################
# Test tags A[i, i] and D
#####################################################################


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [10])
def test_AD(n_cases):
    for p in characteristics():
        sp = space(p)
        for _ in range(n_cases):
            i0 = randint(0, 23)
            scalar = randint(1, p-1) + p * randint(-5, 5)
            c0 = Cocode([i0])
            v0 = GcVector([i0])
            ref = sp('A', i0, i0)
            for j0 in [i0, c0, v0]:
                for j1 in [i0, c0, v0]:
                    assert sp('A', j0, j1) ==  ref
                    mmv = sp()
                    mmv['A', j0, j1] =  scalar
                    assert mmv == scalar * ref
                    assert mmv['A', j0, j1] == scalar % p
                d1 = sp('D', j0)
                assert sp('D', j0) ==  ref
                mmv = sp()
                mmv['D', j0] =  scalar
                assert mmv == scalar * ref
                assert mmv['D', j0] == scalar % p



#####################################################################
# Test tags A, B, and C 
#####################################################################


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [20])
def test_ABC(n_cases):
    for p in characteristics():
        sp = space(p)
        for _ in range(n_cases):
            i1 = i0 = randint(0, 23)
            while i1 == i0:
                i1 = randint(0, 23)
            scalar = randint(1, p-1) + p * randint(-5, 5)
            for tag in "ABC":
                c0 = Cocode([i0])
                c1 = Cocode([i1])
                v0 = GcVector([i0])
                v1 = GcVector([i1])
                ref = sp(tag, i0, i1)
                for j0 in [i0, c0, v0]:
                    for j1 in [i1, c1, v1]:
                        assert sp(tag, j0, j1) ==  ref
                        mmv = sp()
                        mmv[tag, j0, j1] =  scalar
                        assert mmv == scalar * ref
                        assert mmv[tag, j0, j1] == scalar % p
                    

#####################################################################
# Test tag T
#####################################################################


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [20])
def test_T(n_cases):
    for p in characteristics():
        sp = space(p)
        for _ in range(n_cases):            
            o = randint(0, 758)
            so = randint(0, 63)
            coc = Cocode(SubOctad(o, so))
            v = GcVector (coc.syndrome(randint(0,23)))
            sgn = randint(0, 1)
            pl = PLoopZ(sgn) * Octad(o)
            gc = GCode(pl)
            scalar = randint(1, p-1) + p * randint(-5, 5)
            assert gc == GCode(pl)
            ref = (-1)**sgn * sp('T', o, so)
            for sign, d in ((sgn,o), (sgn,gc), (0, pl)):
                for par2 in (so, coc, v):
                    assert (-1)**sign * sp('T', d, par2) == ref
                    mmv = sp()
                    mmv['T', d, par2] = (-1)**sign * scalar
                    assert mmv == scalar * ref
                    signed_scalar = (-1)**sign * scalar % p
                    assert mmv['T', d, par2] == signed_scalar


#####################################################################
# Test tags X, Y, and Z
#####################################################################


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [20])
def test_XYZ(n_cases):
    for p in characteristics():
        sp = space(p)
        for _ in range(n_cases):
            d = randint(0, 0x1fff)
            i = randint(0, 23)
            c = Cocode([i])
            v = GcVector([i])
            pl = PLoop(d)
            gc = GCode(d & 0xfff)
            d0 = d & 0x7ff
            scalar = randint(1, p-1) + p * randint(-5, 5)
            for tag in "XYZ":
                s2 = s1 = d >> 12
                if tag == "Y": s1 ^= d >> 11
                ref = (-1)**s1 * sp(tag, d0, i)
                for sign, dd in ((s1, d0), (s2, gc), (0, pl)):
                    for j in [i, c, v]:
                        assert (-1)**sign * sp(tag, dd, j) == ref
                        mmv = sp()
                        mmv[tag, dd, j] = (-1)**sign * scalar
                        assert mmv == scalar * ref
                        signed_scalar = (-1)**sign * scalar % p
                        assert mmv[tag, dd, j] == signed_scalar



 



from __future__ import absolute_import, division, print_function
#from __future__ import  unicode_literals



import sys
import os
import itertools
from collections import defaultdict, Counter
from numbers import Integral
import numpy as np
import subprocess
import shutil

import pytest
 
from mmgroup.dev.config import INT_BITS
from mmgroup.dev.hadamard.hadamard_codegen import HadamardMatrixCode
from mmgroup.dev.hadamard.hadamard_t import HadamardOpT64
from mmgroup.dev.hadamard.hadamard_t import HadamardOpT3
from mmgroup.dev.hadamard.hadamard_xi import HadamardOpXi64
from mmgroup.dev.hadamard.hadamard_xi import HadamardOpXi16

from mmgroup.tests import test_hadamard
from mmgroup.tests.tempdir import make_temp_dir, kill_temp_dir
from mmgroup.tests.tempdir import temp_dir_name


FILE_NAME = "test_hadamard_op"
C_FILE_NAME = FILE_NAME + ".c"
H_FILE_NAME = FILE_NAME + ".h"
PXD_FILE_NAME = FILE_NAME + ".pxd"
PXD_MODULE_NAME =  "test_hadamard_op"
PYX_FILE_NAME =    "test_hadamard.pyx"

################################################################
# Defining the names of the test functions
################################################################

SIZES = [2, 4, 6]
PRIMES = [3, 7, 15,127, 255]

def hadamard_function_name(p, n_qubits):
    return "mod%d_hadamard%d_test" % (p, 1 << n_qubits)

def xch_parity_function_name(p):
    return "mod%d_xch_parity64_test" % p

def op_t64_function_name(p):
    return "mod%d_op_t64_test" % p
  
def op_t3_function_name(p):
    return "mod%d_op_t3_test" % p

def op_xi64_function_name(p):
    return "mod%d_op_xi64_test" % p

def op_xi16_function_name(p):
    return "mod%d_op_xi16_test" % p

def iter_function_names():
    for q in SIZES:
        for p in PRIMES:
             yield hadamard_function_name(p, q)
    for p in PRIMES:
         yield  xch_parity_function_name(p)
    for p in PRIMES:
         yield  op_t64_function_name(p)
    for p in PRIMES:
         yield  op_t3_function_name(p)
    for p in PRIMES:
         yield  op_xi64_function_name(p)
    for p in PRIMES:
         yield  op_xi16_function_name(p)



################################################################
# Coding test routines in C file
################################################################


def code_hadamard_test(p, n_qubits):
    array_name = "p_data"
    function_name = hadamard_function_name(p, n_qubits)
    s =  """

void %s(uint_mmv_t *%s)
{
""" % (function_name, array_name) 

    cc =  HadamardMatrixCode(p, n_qubits)
    cc.load_vector_direct(array_name)
    cc.hadamard_op()
    cc.mul_pwr2(-n_qubits >> 1)
    cc.store_vector_direct(array_name)
    cc.comment_statistics()
    return  s + cc.generate() +  "}\n"


def code_xch_parity_test(p):
    array_name = "p_data"
    function_name = xch_parity_function_name(p)
    s =  """

void %s(uint_mmv_t *%s)
{
""" % (function_name, array_name) 

    cc = HadamardOpT64(p, 6)
    cc.load_vector_direct(array_name)
    cc.swap_parities()
    cc.store_vector_direct(array_name)
    cc.comment_statistics()
    return  s + cc.generate() +  "}\n"



def code_op_t64_test(p):
    array_name = "p_data"
    function_name = op_t64_function_name(p)
    cc = HadamardOpT64(p, 6)
    s =  """

void %s(uint_mmv_t *%s)
{
uint_mmv_t exp1 = -(%s[%d] & 1ULL);
""" % (function_name, array_name, array_name, cc.V64_INTS) 

    s += cc.make_code(array_name, "exp1", array_name)
    return  s +   "}\n"



def code_op_t3_test(p):
    cc = HadamardOpT3(p)
    d = {
        "array_name" : "p_data",
        "function_name" : op_t3_function_name(p),
        "a_len" : 24 * cc.V24_INTS,
    }
    s =  """

void {function_name}(uint_mmv_t *{array_name})
""".format(**d)  # 1
    s += "{\n"
    s += """
uint_mmv_t exp1 = -({array_name}[3] & 1ULL);
uint_mmv_t a[3 * {a_len}];
a[0] = {array_name}[0];
a[{a_len}] = {array_name}[1];
a[2*{a_len}] = {array_name}[2];
""".format(**d)  # 2
    s += cc.make_code("a", "exp1", "a")
    s += """
{array_name}[0] = a[0];
{array_name}[1] = a[{a_len}];
{array_name}[2] = a[2*{a_len}];
""".format(**d) 
    return  s +   "}\n"






def code_op_xi64_test(p):
    array_name = "p_data"
    function_name = op_xi64_function_name(p)
    cc = HadamardOpXi64(p)
    data = ",\n".join(map(cc.hex, cc.exp_table() + [0] * 8))
    s =  """

void %s(uint_mmv_t *%s)
{
static uint_mmv_t mask0[16] = {
   %s
}; 
int index = (%s[%d]) %% 3;
uint_mmv_t *p_mask = mask0 + (index == 2 ? 8 : index);
""" % (function_name, array_name, data, array_name, 16*cc.V24_INTS) 
    #s += 'printf("OFS = %d\\n", (int)(index));\n' 
    s += cc.make_code(array_name, "p_mask", array_name)
    return  s +   "}\n"




def code_op_xi16_test(p):
    array_name = "p_data"
    function_name = op_xi16_function_name(p)
    cc = HadamardOpXi16(p)
    s =  """

void %s(uint_mmv_t *%s)
{
uint_mmv_t exp1 = -(%s[%d] & 1ULL);
""" % (function_name, array_name, array_name, 4*cc.V24_INTS) 

    s += cc.make_code(array_name, "exp1", array_name)
    return  s +   "}\n"


################################################################
# Creating the .c, .h, .pyd, and .pyx file
################################################################


def generate_test_c_file():
    with open(C_FILE_NAME, "wt") as f:
        f.write("""
#include <stdint.h>
typedef uint%d_t uint_mmv_t;
#include "%s"

"""  % (INT_BITS, H_FILE_NAME))
        for q in SIZES:
            for p in PRIMES:
                 f.write(code_hadamard_test(p, q))
        for p in PRIMES:
            f.write(code_xch_parity_test(p))
        for p in PRIMES:
            f.write(code_op_t64_test(p))
        for p in PRIMES:
            f.write(code_op_t3_test(p))
        for p in PRIMES:
            f.write(code_op_xi64_test(p))
        for p in PRIMES:
            f.write(code_op_xi16_test(p))

def generate_test_h_file():
    with open(H_FILE_NAME, "wt") as f:
        f.write("""
typedef uint%d_t uint_mmv_t;

"""  % INT_BITS)
        for f_name in iter_function_names():
            f.write("void %s(uint_mmv_t*);\n" % f_name)





def generate_test_pxd_file():
    with open(PXD_FILE_NAME, "wt") as f:
        f.write("""# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint%d_t as uint_mmv_t

cdef extern from "%s":

"""  % (INT_BITS, "test_hadamard_op.h" ))
        for f_name in iter_function_names():
            f.write("    void %s(uint_mmv_t*)\n" % f_name)



def generate_test_pyx_file():
    with open(PYX_FILE_NAME, "wt") as f:
        f.write("""# cython: language_level=3

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from libc.stdint cimport uint%d_t as uint_mmv_t

from mmgroup.tests.test_hadamard.temp cimport %s


"""  % (INT_BITS, PXD_MODULE_NAME ))
        for f_name in iter_function_names():
            f.write("""
def %s(v):
    cdef uint_mmv_t[::1] pv = v
    %s.%s(&pv[0])

""" % (f_name, PXD_MODULE_NAME, f_name))





################################################################
# Creating the python library with cython
################################################################




def cythonize():
        """Build the python extension Hadamard64Sub

        This is one of the few places where OS-specific code is required
        """
        if os.name in ["nt"]:
            compilers = ["msvc", ""]
        else:
            compilers = [""]
        for compiler in compilers:
            try:
                command = [sys.executable, 
                   os.path.join("..", "setup_test_hadamard.py"),
                     "build_ext", "--inplace",
                ]
                if compiler: 
                    command.append("-c" + compiler)
                if subprocess.call(command) == 0:
                    return
            except FileNotFoundError:
                pass
        raise ValueError("setup has failed")

def cleanup():
    for fname in [
             C_FILE_NAME, H_FILE_NAME, PXD_FILE_NAME, PYX_FILE_NAME,
             "test_hadamard.c",
        ]: 
        try:
            os.remove(fname)
        except:
            pass
    try:
        shutil.rmtree("build")
    except:
        pass
        

################################################################
# Main program
################################################################


def build():
    old_dir = os.getcwd()
    temp_dir = make_temp_dir(test_hadamard)
    os.chdir(temp_dir)
    generate_test_c_file()
    generate_test_h_file()
    generate_test_pxd_file()
    generate_test_pyx_file()
    cythonize()
    cleanup()
    os.chdir(old_dir)


def kill():
    kill_temp_dir(temp_dir_name(test_hadamard))

 


r"""We deal with the Golay code and its cocode.

Let :math:`\tilde{\Omega}` be the set of integers
:math:`\{ i \in \mathbb{Z} \mid 0 \leq i < 24\}` of size 
:math:`24` and construct the vector space
:math:`\mathcal{V} = \mathbb{F}_2^{24}` as 
:math:`\prod_{i\in \tilde{\Omega}} \mathbb{F}_2`. In this 
document elements of :math:`\mathcal{V}` are called *bit vectors*.
We represent bit vectors as instances of class |GcVector|.

We identify  the power set of :math:`\tilde{\Omega}` with the
vector space :math:`\mathcal{V} = \mathbb{F}_2^{24}` by mapping 
each subset of :math:`\tilde{\Omega}` to its characteristic 
function, which is a bit vector in :math:`\mathcal{V}`.  So we may 
talk about union and intersection of bit vectors in a natural way 
and use the python operators ``|`` and ``&`` for these operations 
as usual. We use the  ``+`` operator for vector addition of bit 
vectors. Bit vectors are numbered from ``0`` to ``0xffffff``, with 
bit ``i`` of that number corresponding to the ``i``-th unit vector.

The Golay code :math:`\mathcal{C}` is a :math:`12`-dimensional 
linear subspace of :math:`\mathcal{V}` such that a nonzero 
vector in :math:`\mathcal{C}` has weight at least  :math:`8`.
This characterizes the Golay code up to permutation. In this
package we chose a specific basis of the Golay code that
satisfies the requirements in :cite:`Seysen20`, see
:ref:`basis-golay-label` for details. We represent Golay
code words as instances of class |GCode|. 
Golay code words of weight :math:`8` and :math:`12` are called 
*octads* and *dodecads*, respectively. 

We stress that  class |GCode| is not a subclass of class |GcVector|.
Instances of these two classes are numbered in a different way.
Golay code words are numbered from ``0`` to ``0xfff`` with bit
``i`` of that number corresponding to the ``i``-th basis vector
of the Golay code. See :ref:`basis-golay-label` for the chosen 
basis of the Golay code. 

The Golay cocode :math:`\mathcal{C}^*` is the ``12``-dimensional 
quotient space :math:`\mathbb{F}_2^{24} / \mathcal{C}`. For any
:math:`\delta \in \mathcal{C}^*` the weight of :math:`\delta`
is the weight of its lightest representative in
:math:`\mathbb{F}_2^{24}`.  That weight is at most ``4``.
There is a unique lightest representative  it the weight is
less than ``4``, and there are six mutually disjoint  
lightest representatives of an element of weight  ``4``.
A partition of  :math:`\mathbb{F}_2^{24}` given by such a set of 
disjoint representatives is called a *tetrad*. We represent 
elements of the cocode of the Golay code as instances of class 
|Cocode|. Cocode elements are numbered from ``0`` to ``0xfff`` 
with bit ``i`` of that number corresponding to the ``i``-th basis 
vector of the cocode. Our basis of the cocode is the reciprocal 
basis of the basis of the Golay code, see :ref:`basis-golay-label` 
for details.

The *scalar product* of a Golay code word and a cocode element is 
the parity of the intersection of the Golay code word and any 
representative of the cocode element. If ``g`` is an instance of 
class |GCode| and ``d`` is an instance of class |Cocode| then
``g & d`` is an instance of class |Parity|. Here class |Parity|
models an integer modulo ``2``, which is also considered as the
parity of a bit vector. Standard operations, e.g. ``(-1)**p``,
work as expected for an instance ``p`` of class |Parity|.

The standard function ``len(x)`` returns the (minimal) weight 
``|x|`` of an instance ``x`` of class |GcVector|, |GCode| or 
|Cocode|. For instances ``g1, g2, g3`` of class  |GCode|, the
expression ``g1 & g2`` is an instance of class  |Cocode| and
``g1 & g2 & g3`` an instance of class |Parity|; these values 
have a natural interpretations as bitwise intersections.  
``g1/4`` is a shorthand for ``Parity(len(g1)/4)``; and
``(g1 & g2)/2`` is a shorthand for ``Parity(len(g1 & g2)/2)``. 
These two expressions are called the *power map* and the
*commutator*, respectively, in  :cite:`Asc86`;
and ``g1 & g2 & g3`` is called the *associator* in 
:cite:`Asc86`.

The numbering of the Golay code words and of the cocode elements 
is important for indexing some generators of 
:math:`\mathbb{M}` and also some basis vectors in the
representations  :math:`\rho_p` of :math:`\mathbb{M}`. 
"""

from functools import reduce
from operator import __xor__
from numbers import Integral, Number
from random import randint

try:
    # Try importing the fast C function
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24



from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.parity import Parity
from mmgroup.structures.parse_atoms import ihex



ERR_RAND = "Illegal string for constructing type %s element" 

ERR_DIV4 = "%s object may be divided by 2 or 4 only"
ERR_DIV2 = "%s object may be divided by 2 only"


#######################################################################
# Import derived classed
#######################################################################


import_pending = True

def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending, Cocode, PLoop, PLoopIntersection
    global AutPL, AutPlGroup, XLeech2
    from mmgroup.structures.cocode import Cocode
    from mmgroup.structures.cocode import PLoopIntersection
    from mmgroup.structures.ploop import PLoop
    from mmgroup.structures.autpl import AutPL, AutPlGroup
    from mmgroup.structures.xleech2 import XLeech2
    import_pending = False

#######################################################################
# Auxiliary functions
#######################################################################



def as_vector24(data):
    """Convert list of bit positions to vector in ``GF(2)**24``"""
    try:
        vector = reduce(__xor__, (1 << x for x in data), 0)
    except:
        err = "List of integers >= 0 required for a 24-bit vector"
        raise TypeError(err)
    if vector >= 0x1000000:
        raise ValueError("List entry too large for a 24-bit vector")
    return vector


def str_vector24(v):
    """Return the 24-bit vector ``v`` as a string of binary digits

    Here ``v`` must be an integer with the coefficient of the
    ``i``-th bit coded in the bit with valence ``2**i``.
    Coefficients  ``i >= 24`` are ignored.
    """
    l = ["01"[(v >> i) & 1] for i in range(24)]
    l4 = ("".join(l[i:i+4]) for i in range(0, 24, 4))
    return " ".join(l4)

def str_basis(text, letter, basis):
   s = text + "\n"
   for i, v in enumerate(basis):
       s += "  %3s: %s\n" % (letter + str(i), str_vector24(v)) 
   return s


#######################################################################
# Class GCode 
#######################################################################



class GCode():
    """This class models a code word of the Golay code.

    The Golay code is a binary ``[24, 12, 8]`` code. So there are 
    :math:`2^{12}` code words, each word is :math:`24` bit long, and
    the  Hamming distance between two different code words is at least 
    :math:`8`.

    The :math:`2^{12}` Golay code words are numbered from
    ``0`` to ``0xfff``. Bit ``i`` in that number refers to the
    coefficient of the  ``i``-th basis vector of the Golay code,
    which may be :math:`0` or :math:`1`.


    :param value:

      This is the value of the Golay code word to be returned. 

    :type value: see table below for legal types

    :return: A Golay code vector 
    :rtype:  an instance of class |GCode|

    :raise:
        * TypeError if ``type(value)`` is not in the table given below.
        * ValueError if ``value`` cannot be converted to an
          instance of class  |GCode|.


    Depending on its type parameter **value** is  interpreted as follows:

    .. table:: Legal types for constructor of class ``GCode``
      :widths: 20 80

      ====================== ================================================
      type                   Evaluates to
      ====================== ================================================
      ``int``                Here the code word with number ``value`` is
                             returned.  ``0 <= value < 0x1000`` must hold.
  
      ``list`` of ``int``    A list of integers  ``0 <= i < 24`` is 
                             interpreted as a list of bit positions
                             to be set in a bit vector. The Golay code word
                             corresponding to that bit vector is returned.
                             Up to ``3`` erroneous bits are corrected.

      class |GCode|          A deep copy of parameter ``value`` is returned. 
 
      class |PLoop|          The Parker loop element ``value`` is converted
                             to a Golay code word (by dropping its sign)
                             and returned.

      class |GcVector|       The bit vector ``value`` of type |GcVector| is 
                             converted to a Golay code word and returned. 
                             Up to ``3``  erroneous bits
                             in that vector are corrected.

      class |XLeech2|        The Golay code part of the |XLeech2| object  
                             is converted to a Golay code word and returned. 

       ``str``               Create random element depending on the string
                              | ``'r'``: Create arbitrary Golay code word
      ====================== ================================================


    **Standard operations**
    
    Addition of Golay code words and scalar multiplication (with scalars 
    of type  ``int`` or |Parity|) are done as usual.

    The bitwise and operator ``&`` means bitwise *and* of two code words
    of type  |GCode|. The result of such an operation on two Golay code
    words is a Golay cocode word of type |Cocode|.

    The bitwise ``&`` operator of a Golay code word and a Golay cocode
    word of type |Cocode| is not well defined as a bit vector, but it
    has a well-defined parity. So in this case the result of the 
    ``&`` operator is a parity object of type |Parity|.

    For the bitwise ``&`` operator bit vectors of type |GcVector|, see
    class |GcVector|.

    The   ``~``  operator means bitwise complement  as usual; it
    returns the complemented code word, which is also of type |GCode|.

    **Standard functions**

    ``len(g)`` returns the bit weight of the Golay code word ``g``, i.e.
    the number of bits set in the word ``g``. 


    **Special functions**

    Let ``g1``,  ``g2``, and ``g3`` be Golay code vectors of 
    type |GCode|.
  
    The expression ``g1/4`` is a shorthand for  ``Parity(len(g1)/4)``.
    It returns   ``len(g1)/4``  modulo ``2`` as a parity
    object of type  |Parity|.  
    
    ``GcVector(g1 & g2)`` returns the bitwise intersection of 
    ``g1`` and ``g2`` as a bit vector of type  |GcVector|.

    ``(g1 & g2)/2`` is a shorthand for 
    ``Parity(len(GcVector(g1 & g2))/2)``. It returns the
    halved bit weight of ``g1 & g2`` modulo ``2`` as a parity
    object of type  |Parity|.      
    This value is also called the *commutator* of ``g1`` and ``g2``. 

    ``g1 & g2 & g3`` returns the parity of the bitwise intersection 
    of   ``g1``,  ``g2``, and ``g3`` as a parity object
    of type  |Parity|. This value is also called the
    *associator* of ``g1``, ``g2``, and ``g3``.

    ``abs(g1)`` is a shorthand for ``abs(PLoop(g1))``, see 
    class   |PLoop| for details.

    """
    __slots__ = "value", "bit_list_"
    basis = mat24.basis[12:24]
    parity = 0
    cocode = 0
    sign = 1
    parity_class = True

    def __init__(self, value):
        """Constructor"""
        if import_pending:
            complete_import()
        if isinstance(value, Integral):
            self.value = value & 0x0fff
        elif isinstance(value, GCode):
            self.value = value.value & 0xfff
        elif isinstance(value, GcVector):
            vector = value.value
            vector ^= mat24.syndrome(vector) 
            self.value = mat24.vect_to_gcode(vector)
        elif isinstance(value, XLeech2):
            self.value = (value.value >> 12) & 0x1fff
        elif isinstance(value, str):
           self.value = randint(0, 0xfff)
        else:
            vector = as_vector24(value)
            vector ^= mat24.syndrome(vector) 
            self.value = mat24.vect_to_gcode(vector)       


    def __eq__(self, other):
        return (isinstance(other, GCode)  
            and ((self.value ^ other.value) & 0xfff) == 0)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pos__(self):
        return self
        
    __neg__ = __pos__

    def __and__(self, other):
        if import_pending:
            complete_import()
        if isinstance(other, GCode):
            return PLoopIntersection(self, other)
        elif  isinstance(other, Cocode):
            return Parity(mat24.scalar_prod(self.value, other.value))
        elif isinstance(value, GcVector):
            return  GcVector(mat24.gcode_to_vect(self.value) & other.value)
        else:
            return NotImplemented
            

    def __add__(self, other):
        if (isinstance(other, GCode)):
            return GCode((self.value ^ other.value) & 0xfff)
        elif isinstance(other, GcVector):
            return  GcVector(mat24.gcode_to_vect(self.value) ^ other.value)
        elif other == 0:
            return GCode(self)
        else:
             return NotImplemented


    __radd__ = __add__
    __sub__ = __add__
    __rsub__ = __add__


    def __mul__(self, other):
        if isinstance(other, Integral):
            mask = -(other & 1)
            return GCode(self.value & mask & 0xfff)
        elif isinstance(other, AutPL):
            return GCode(mat24.op_gcode_perm(self.value, other.perm))
        else:
             return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, Integral):
            mask = -(other & 1)
            return GCode(self.value & mask & 0xfff)
        else:
            return NotImplemented

    def __len__(self):
        return mat24.gcode_weight(self.value) << 2

    def __abs__(self):
        return PLoop(self.value & 0xfff)

    def __invert__(self):
        return  GCode(self.value ^ 0x800)

    def __truediv__(self, other):
        if isinstance(other, Integral):
            other = abs(other)
            if other == 1:
                return self
            elif other == 2:
                return Parity(0)
            elif other == 4:
                return Parity(mat24.gcode_weight(self.value) & 1)
            else:
                raise ValueError(ERR_DIV4 % type(self))
        else:
             return NotImplemented

    def __mod__(self, other):
        return Parity(self.parity) % other


    @property
    def ord(self):
        """Return the number of the Golay code word.

        We have ``0 <= i < 0x1000`` for the returned number ``i``.
        """
        return self.value & 0xfff


    @property
    def octad(self):
        """Return number of octad of the Golay code word as an octad.
        
        Complements of octads are also accepted as octads.   We have 
        ``0 <= o < 759`` for the returned number ``o``.

       :raise:
        * ValueError if the Golay code word is not an octad.


        """
        return mat24.gcode_to_octad(self.value & 0xfff, 0)

    @property
    def gcode(self):
        """Return the number of the Golay code word.

        Same as property ``ord``.
        """
        return self.value & 0xfff
        
    @property    
    def vector(self):
        """Return bit vector corresponding to the Golay code word as an int.

        We have ``0 <= v < 0x1000000`` for the returned number ``v``.
        Bit ``i`` of the number ``v`` is the  ``i``-th coordinate
        of the corresponding bit vector.
        """
        return mat24.gcode_to_vect(self.value)
        
    @property    
    def bits(self):
        """Return the Golay code word as a ``24``-bit vector.

        The function returns a list with  ``24`` entries equal to
        ``0`` or ``1``.
        """
        try:
            return self.bit_vector
        except:
            v = mat24.gcode_to_vect(self.value)
            self.bit_vector = ([(v >> i) & 1 for i in range(24)]) 
            return self.bit_vector

    @property    
    def bit_list(self):
        """Return the ordered list of positions of bits being set"""
        try:
            return self.bit_list_
        except AttributeError:
            self.bit_list_ = mat24.gcode_to_bit_list(self.value)
            return self.bit_list_
             

    def split(self):
        """Split the Golay code word ``g``. 

        :return: A triple ``(0, eo, v)`` with  ``g  =  Omega * eo + v``.

        Here ``Omega`` is the Golay code word containing  one bits only,
        ``eo`` is ``0`` or ``1``, and ``v`` is a Golay code word 
        of type |GCode| with ``0 <= v1.ord < 0x800``.

        This method is for compatibility with the corresponding method in
        class |PLoop|.
        """
        return 0, (v >> 11) & 1, GCode(v & 0x7ff)
 
    def split_octad(self):
        """Split an octad from the Golay code word ``g``. 

        :return: A triple ``(0, eo, o)`` with  ``g  =  Omega * eo + o``.

        Here ``Omega`` is the Golay code word containing  one bits only,
        ``eo`` is ``0`` or ``1``, and ``v`` is a Golay code word 
        of type |GCode| which is either the zero code word or an
        octad, i.e. a code word of weight ``8``.

        This method is for compatibility with the corresponding method 
        in class |PLoop|.

        :raise:
          * ValueError if this is not possible.
        """
        v = self.value
        eo, w = 0, mat24.gcode_weight(v)
        if mat24.gcode_weight(v) > 3:
            v, eo, w = v ^ 0x800, 1, 6 - w
        if w <= 2:
            return 0, eo, GCode(v & 0xfff)
        raise ValueError("Cannot convert Golay code word to octad")
        
    def theta(self, g2 = None):
        """Return cocycle of Golay code words.

        The cocycle ``theta`` maps a pair of Golay code words 
        to an integer modulo ``2``. 
        It is linear in its second argument, so it
        may also be considered as a mapping from the Golay code
        to the cocode of the Golay code.


        :param g2:  ``None`` (default) or another Golay code word of
                    type |GCode|.
        :returns:
            * If ``g2`` is a code word of type |GCode|, we return
              the value ``g1.theta(g2) = theta(g1, g2)`` 
              as a |Parity| object.
            * If ``g2`` is ``None`` (default), we return the value
              ``g1.theta() = theta(g1)``  as a |Cocode| object.
              Note that  ``g1.theta(g2) == g1.theta() & g2 .``

        The importance of the ``theta`` function comes from the
        fact that the multiplication of elements of the Parker loop
        is based on the cocycle. We embed the set of Golay code words  
        into  the set of positive Parker loop elements, which are 
        instances of class |PLoop|. 
 
        Let ``g1`` and ``g2`` be Golay code words of type |GCode|.
        Then  ``PLoop(g1)`` and ``PLoop(g2)`` are the corresponding
        positive Parker loop elements,  and  ``g1.theta(g2)`` is an 
        integer  modulo ``2`` of type |Parity|. We have:  
       
        ``PLoop(g1) * PLoop(g2) == (-1)**g1.theta(g2) * PLoop(g1 + g2) .``
        """
        th = mat24.ploop_theta(self.value) 
        if g2 == None:
            complete_import()
            return Cocode(th)
        if isinstance(g2, GCode):
            return Parity(mat24.scalar_prod(g2.value, th))
        err = "Types %s is illegal for method theta()"
        raise TypeError(err % type(g2))

    def str(self):
        return "<GCode_%s>" % ihex(self.value & 0xfff, 3)
    __repr__ = str

    @classmethod
    def str_basis(cls):
       b = "Golay code basis g_i[0,...,23], i = 0,...,11"
       return str_basis(b, "g", cls.basis)

    @classmethod
    def show_basis(cls):
        print(cls.str_basis())







#######################################################################
# Class GcVector
#######################################################################

class GcVector:
    """Models a 24-bit vector in the underlying space of the Golay code.

    The :math:`2^{24}` bit vectors are numbered from
    ``0`` to ``0xffffff``. Bit ``i`` in that number refers to the
    coefficient of the  ``i``-th basis vector of the vector space,
    which may be :math:`0` or :math:`1`.

    :param value:

      This is the value of the bit vector to be returned. 

    :type value: see table below for legal types

    :return: A bit vector of 24 bit length.
    :rtype:  an instance of class |GcVector|

    :raise:
        * TypeError if ``type(value)`` is not in the table given above.
        * ValueError if ``value`` cannot be converted to an
          instance of class  |GcVector|.

    Depending on its type parameter **value** is  interpreted as follows:

    .. table:: Legal types for constructor of class ``GcVector``
      :widths: 20 80

      ===================== ================================================
      type                  Evaluates to
      ===================== ================================================
      ``int``               Here the bit vector with number ``value`` is
                            returned. 
                            ``0 <= value < 0x1000000`` must hold.

      ``list`` of ``int``   A list of integers  ``0 <= i < 24`` is 
                            interpreted as a list of bit positions
                            to be set in a bit vector. That bit vector 
                            is returned.

       class |GCode|        The bit vector corresponding to the Golay 
                            code word ``value`` is returned. 

       class |PLoop|        The Parker loop element ``value`` is converted
                            to a Golay code word (by dropping its sign)
                            and then to a bit vector. That bit vector is
                            returned.

       class |GcVector|     A deep copy of parameter ``value`` is returned.


      ``str``               Create random element depending on the string
                             | ``'r'``: Create arbitrary bit vector

      ===================== ================================================


    **Standard operations**
    
    Addition of bit vectors and scalar multiplication (with scalars of 
    type  ``int`` or |Parity|) are done as usual. The sum of a bit
    vector and a Golay code word (of type |GCode|) is a bit vector.
    The sum of a bit vector and a cocode word (of type |Cocode|) is a 
    cocode word of type |Cocode|.

    The bitwise and operators ``&``, ``|``, and ``~`` operate on two
    bit vectors as expected. If the other operand is a Golay code
    vector (of type |GCode|) then it is converted to a bit vector. 

    **Standard functions**

    ``len(v)`` returns the bit weight of the bit vector ``v``, i.e.
    the number of bits set in the vector ``v``. 
 
    """
    __slots__ = "value", "bit_list_"
    parity_class = True
    sign = 1

    def __init__(self, value):
        if import_pending:
            complete_import()
        if isinstance(value, Integral):
            self.value = value & 0xffffff
        elif isinstance(value, GCode):
            self.value = mat24.gcode_to_vect(value.value)
        elif isinstance(value, GcVector):
            self.value =  value.value
        elif isinstance(value, PLoopIntersection):
            self.value = (mat24.gcode_to_vect(value.v1) & 
                mat24.gcode_to_vect(value.v2) & 0xffffff)
        elif isinstance(value, str):
            self.value = randint(0, 0xffffff)
            if 'e' in value and not 'o' in value:
                if mat24.bw24(self.value) & 1:
                    self.value ^= 1 << randint(0, 23)
            if 'o' in value and not 'e' in value:
                if not mat24.bw24(self.value) & 1:
                    self.value ^= 1 << randint(0, 23)
        else:
            self.value = as_vector24(value)
      

    def __add__(self, other):
        if isinstance(other, GcVector):
            return GcVector(self.value ^ other.value)
        elif isinstance(other, GCode):
            return GCode(self.value ^ mat24.gcode_to_vect(other.value))
        elif other == 0:
            return self
        else:
            if import_pending:
                complete_import()
            return Cocode(self).__add__(other)

    __radd__ = __add__
    __sub__ = __add__
    __rsub__ = __add__

    def __and__(self, other):
        if isinstance(other, GcVector):
            return GcVector(self.value & other.value)
        elif isinstance(other, GCode):
            return GCode(self.value & mat24.gcode_to_vect(other.value))
        else:
            if import_pending:
                complete_import()
            return Cocode(self).__and__(other)

    __rand__ = __and__

    def __or__(self, other):
        if isinstance(other, GcVector):
            return GcVector(self.value | other.value)
        elif isinstance(other, GCode):
            return GCode(self.value | mat24.gcode_to_vect(other.value))
        else:
            return NotImplemented

    __ror__ = __and__

    def __invert__(self):
        return GcVector(self.value ^ 0xffffff)

    def __eq__(self, other):
        return (isinstance(other, GcVector)  
            and ((self.value ^ other.value) & 0xffffff) == 0)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pos__(self):
        return self

    __neg__ = __pos__

    def __mul__(self, other):
        if isinstance(other, Integral):
             return GcVector(self.value & -(other & 1))
        elif isinstance(other, AutPL):
            return GcVector(mat24.op_vect_perm(self.value, other.perm))
        else:
            return NotImplemented
      
    def __rmul__(self, other):
        if isinstance(other, Integral):
             return GcVector(self.value & -(other & 1))
        else:
            return NotImplemented
     
    def __truediv__(self, other):
        if isinstance(other, Integral):
            other = abs(other)
            if other == 1:
                return self
            w = mat24.bw24(self.value)
            if other == 2 and  w & 1 == 0:
                return Parity(w >> 1)
            elif other == 4 and  w & 3 == 0:
                return Parity(w >> 2)
            elif other in (2,4):
                err = "Don't know weight of Omega vector / %d"
                raise ValueError(err % other)
            else:
                raise TypeError(ERR_DIV4 % self.__class__)
        else:
            return NotImplemented

    def __mod__(self, other):
        return Parity(self.parity) % other


    def __len__(self):
        return mat24.bw24(self.value)

    @property
    def ord(self):
        return self.value & 0xffffff

    @property
    def parity(self):
        return mat24.bw24(self.value) & 1
        
    @property
    def cocode(self):
        """Return number of the cocode word corresponding to the vector.
       
        """
        return mat24.vect_to_cocode(self.value)

    @property
    def octad(self):
        """If the bit vector is an octad, return the number of that octad
        
        Complements of octads are also accepted as octads.

        :raise: 
            * ValueError if the bit vector is not an octad.
        """
        return mat24.vect_to_octad(self.value, 0)

    @property
    def gcode(self):
        """Return number of Golay code word corresponding to the bit vector 
        
        :raise: 
             * ValueError if the bit vector is not an Golay code word.
        """
        return mat24.vect_to_gcode(self.value)

    @property
    def ord(self):
        """Return the number of the bit vector.

        We have ``0 <= i < 0x1000000`` for the returned number ``i``.
        """
        return self.value & 0xffffff

        
    @property    
    def vector(self):
        """Return the number of the bit vector.

        We have ``0 <= i < 0x1000000`` for the returned number ``i``.

        Same as property ``ord``.
        """
        return  self.value & 0xffffff
        
    @property    
    def bits(self):
        """Return the Golay code word as a ``24``-bit vector.

        Returns a list with  ``24`` entries equal to
        ``0`` or ``1``.
        """
        try:
            return self.bit_vector
        except:
            self.bit_vector = ([(self.value >> i) & 1 for i in range(24)]) 
            return self.bit_vector

    @property    
    def bit_list(self):
        """Return the ordered list of positions of bits being set"""
        try:
            return self.bit_list_
        except AttributeError:
            self.bit_list_ = [i for i in range(24) if self.value & (1<<i)]
            return self.bit_list_

    def syndrome(self, i=None):
        """Return syndrome of cocode element as a bit vector.

        :param i: ``None`` (default) or an integer ``0 <= i < 24`` 
                  used to select a syndrome.
                 
        :return: Syndrome of a Golay cocode element as a bit
                 vector of type |GcVector|.

        :raise:
           *  ValueError if argument ``i`` is ``None`` and the
              syndrome is not unique.

        ``v.syndrome(i)`` is equivalent to ``Cocode(v).syndrome(i)``.
        """
        if i is None: i = 24
        return GcVector(mat24.syndrome(self.value, i))

    def syndrome_list(self, i=None):
        """Return syndrome of cocode element as list of bit positions.

        :param i: ``None`` (default) or an integer ``0 <= i < 24`` 
                  used to select a syndrome.
                 
        :return: Syndrome of a Golay cocode element as a list
                 of at most four bit positions.

        :raise:
           *  ValueError if argument ``i`` is ``None`` and the
              syndrome is not unique.

        The syndrome of the Golay cocode element is calculated
        in the same way as in method **syndrome**. But here
        the result is returned as an ordered list of bit positions
        corresponding to the bits which are set in the syndrome.

        ``v.syndrome_list(i)`` is equivalent to 
        ``Cocode(v).syndrome_list(i)``.
        """
        if i is None: i = 24
        syn = mat24.syndrome(self.value, i)
        return [j for j in range(24) if (1 << j) & syn]



    def str(self):
        return "<GcVector_%s>" % ihex(self.value & 0xffffff, 6)
    __repr__ = str

        


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__
from numbers import Integral

import numpy as np
import pytest

from mmgroup import MM0
from mmgroup.structures.qs_matrix import qs_pauli_matrix
from mmgroup.structures.xsp2_co1 import Xsp2_Co1, str_leech3
from mmgroup.structures.xsp2_co1 import get_error_pool
from mmgroup.structures.autpl import AutPL
from mmgroup.mat24 import MAT24_ORDER, ploop_theta
from mmgroup.tests.spaces.clifford_space import Space_ZY
from mmgroup.generators import gen_leech2to3_short, gen_leech3to2_short
from mmgroup.generators import gen_leech2_op_word
from mmgroup.clifford12 import xsp2co1_find_chain_short_3
from mmgroup.clifford12 import xsp2co1_chain_short_3


MMSpace3 = Space_ZY.mmspace
MMGroup3 = MM0 # MMSpace3.group



STD_V3  = 0x8000004





#####################################################################
# Creating test vectors
#####################################################################


def rand_tuple(tag):
    if tag in "BC":
        i0 = i1 = randint(0, 23)
        while i1 == i0:
            i1 = randint(0, 23)
        return tag, i0, i1
    if tag == "T":
        return tag, randint(0, 758), randint(0, 63),
    if tag in "XY":
        return tag, randint(0, 2047), randint(0, 23),
    if tag == "d":
        return tag, randint(0, 0xfff)
    if tag == "p":
        return tag, randint(0,  MAT24_ORDER - 1)
    if tag in  "xy":
        return tag, randint(0, 0x1fff)
    if tag ==  "l":
        return tag, randint(1, 2)
    raise ValueError("Illegal tag %s for group or vector" % tag)
        

def rand_element(s):
    return [rand_tuple(tag) for tag in s]

def create_test_vectors():
    vector_data = [ 
      (0x001, (1, "B", 2, 3)),
      (0x001, (1, "C", 0, 16)),
      (0xabf, (1, "C", 15, 19)),
      (0x001, (1, "T", 643, 51)),
      (0x001, (1, "T", 199, 7)),
      (0x001, (1, "X", 0, 23)),
      (0x001, (1, "X", 137, 23)),
      (0x001, (1, "X", 1897, 1)),
    ]
    group_data = [
      [('x', 0x1f24), ('d', 0xf75)],
      [('x', 0x124), ('d', 0x555)],
      [('d', 0x124)],
      [('d', 0x800)],
      [('p', 187654344)],
      [('d', 0xd79), ('p', 205334671)],
      [('p', 205334671), ('d', 0xd79)],
      [('d', 0xd79), ('x', 0x1123)],
      [('y', 0x1d79)],
      [('y', 0x586)],
      [('l', 1)],
      [('l', 2)],
    ]
    for i in range(1):
        p = {0:2, 1:3, 2:0, 3:1, 4:4, 6:6, 8:8}
        #print("AutPL", [hex(x) for x in AutPL(p).rep])
        yield 256, (1, "B", 2, 3),  [("p", p)]
    for x4096, x24 in vector_data:
        for g in group_data:
            yield x4096, x24, g
    group_elements = "lydpdpxdpylx"
    #group_elements = "ydpdpxdpyx"
    for x in "BCTX":
        for j in range(50):
            sign = -1**j
            d = randint(0, 0xfff)
            t = rand_tuple(x)
            yield d, (sign,) + t,  rand_element(group_elements)    


#####################################################################
# Simulate the mapping of Leech lattice vectors modulo 3
#####################################################################



def tuple_to_leech3(sign, tag, i0, i1):
    x2 = MMSpace3.index_to_short_mod2(tag, i0, i1)
    x3 = gen_leech2to3_short(x2)
    if sign == -1:
        x3 ^= 0xffffffffffff
    return x3
    


def short3_abs_equal(x1, x2):
    x1, x2 = int(x1), int(x2)
    return not bool((x1 ^ (x1 >> 24)  ^ x2 ^ (x2 >> 24)) & 0xffffff)

def map_v3(v, g, expected = None, verbose = 1):
    src = np.zeros(3, dtype = np.uint64)
    dest = np.zeros(3, dtype = np.uint64)
    src[0] = 0x8000004
    dest[0] = g.short3
    src[2] = v if isinstance(v, Integral) else tuple_to_leech3(*v)
    src[1] = xsp2co1_find_chain_short_3(src[0], src[2])
    qstate = g.qs
    qstate_base = qstate.copy().gate_h(0x800800)
    xsp2co1_chain_short_3(qstate, src, dest)
    ok = dest[-1] != 0
    if not expected is None:
        ok = ok and short3_abs_equal(dest[-1], expected)
    if verbose or not ok:
        print("Map a short Leech vector v by an element g of Xsp2_Co1:")
        if not isinstance(v, Integral):
            print("v = %s = %s" % (v, str_leech3(src[2])))
        else:
            print("v = %s" % str_leech3(src[2]))
        print("g =", str(g))
        for i in range(3):
            print(" ",  str_leech3(src[i]), "->", str_leech3(dest[i]))
        for i in range(3):
            v2 =  gen_leech3to2_short(src[i])       
            v2c = qstate_base.pauli_conjugate(v2);
            v3c = gen_leech2_short_to3(v2c)
            print("   %s -> 0x%06x -> 0x%06x -> %s" %
                (str_leech3(src[i]), v2, v2c, str_leech3(v3c)))
            print("    Debug data pool:\n    ", 
                    [hex(x) for x in get_error_pool(15)])
        print("")
        if not expected == None:
            print("Expected result: %s" % str_leech3(expected))
        print("Obtained result: %s\n" % str_leech3(dest[-1]))
        if not ok:
            ERR = "Mapping of Leech vector failed"
            raise ValueError(ERR)
    return dest[-1]


#####################################################################
# Test multiplication of vectors with elements of G_{x1}
#####################################################################


@pytest.mark.xsp2co1
def test_vector(verbose = 0):
    for ntest, (x4096, x24, g) in enumerate(create_test_vectors()):
        if verbose:
            print("\nTEST %s" % (ntest+1))
            print("vector =", x4096, x24)
        vm = x24[0] * Space_ZY.unit(x4096, x24[1:])
        vm_old = vm.copy()
        if verbose:
            print("vm =", vm.as_tuples())
            print("vm =", vm)
            vm.dump()
        v3 = vm.as_mmspace_vector() 
        if verbose:
            print("g =", g)
        g3 = MMGroup3(g)
        try:
            gm = Xsp2_Co1(g)
        except ValueError:
            print("\nError in constructing group element g")
            print("Debug data pool:\n", 
                    [hex(x) for x in get_error_pool(15)])
            raise
        gm_old = gm
        if verbose:
            print("gm = ", gm)
            print("g3 = ", g3)
        try:
            wm = vm * gm
            if verbose:
                print("w = vm * gm = ", wm)
        except ValueError:
            print("Debug data pool:\n", 
                    [hex(x) for x in get_error_pool(15)])
            map_v3(x24, gm, wm.short3, verbose = 1)
            raise
        w3_op =  wm.as_mmspace_vector()
        w3_mult =  v3 * g3
        if verbose:
            print("w3_op =", w3_op)
            print("w3_op data =")
            wm.dump()
            print("w3_mult =", w3_mult)
        ok =  w3_op == w3_mult 
        assert vm == vm_old
        assert gm == gm_old
        if not ok:
            print("\nError in TEST %s" % (ntest+1))
            print("vector =", x4096, x24)
            print("v =", hex(x4096), "(x)", x24)
            print("rep of v:", vm)
            print("v =", v3)
            print("g = ", g)
            print("rep of g:", gm)
            print("Output w = v * g\nexpected:", w3_mult)
            print("obtained:", w3_op)
            print("rep:", wm)
            map_v3(x24, gm, verbose = 1)
            if not ok:
               ERR = "Vector multiplication test in group G_{x0} failed"
               raise ValueError(ERR)
            
            
            
            
            
#####################################################################
# Test conjugation with some elements of group G_{x1}
#####################################################################




   
def ref_conjugate(x, elem):
    elem_l = Xsp2_Co1(elem)
    mat_l = elem_l.qs.gate_h(0x800800)
    mat_x = qs_pauli_matrix(12, x)
    mat_res = mat_l.inv() @ mat_x @ mat_l
    return mat_res.pauli_vector()
   

   

TAGS = {'l':6, 'd':1, 'p':2, 'x':3, 'y':4 }

def conjugate_gen_xi(x, elem):
    data = [(TAGS[tag] << 28) + (d & 0xfffffff) for tag, d in elem]
    a = np.array(data, dtype = np.uint32)
    return gen_leech2_op_word(x, a, len(a))


def conjugate_group_testdata():
    max_tag = {'x': 0x1fff, 'y': 0x1fff, 'd': 0xfff, 'p': MAT24_ORDER-1}
    data = [
        [("l",1)],
        [("l",2)],
    ]
    for d in data: 
        yield d
    for tag in "pydx":
        for i in range(30):
            yield [(tag, randint(0, max_tag[tag]))]
    for i in range(30):
        yield [('d', randint(0, 0xfff)), 
              ('p', randint(0, MAT24_ORDER-1))]


def conjugate_vector_testdata():
    for i in range(25):
        yield (1 << i) 
    for n in range(20):
        yield randint(0, 0x1ffffff)




@pytest.mark.xsp2co1
def test_conjugate_mat24xi(verbose = 0):
    for g in conjugate_group_testdata():
        for x in conjugate_vector_testdata():
            xc_ref = ref_conjugate(x, g)
            xc  = conjugate_gen_xi(x, g)
            ok = xc == xc_ref
            #ok_upto_sign = ok or ((xc ^ xc_ref) & ~0x1800000 )== 0
            if verbose or not ok:
                g_hex = [(t, hex(x)) for t,x in g]
                print("conjugate x= 0x%06x with  %s" % (x, g_hex))
                if not ok:
                    print("expected: 0x%06x" % xc_ref) 
                print("obtained: 0x%06x" % xc) 
                if not ok:
                    ERR = "Error in conjugating with element l"
                    raise ValueError(ERR)



     

#####################################################################
# Test multiplication ind inversion of elements of G_{x1}
#####################################################################
           

def create_test_elements():
    for i in range(50):
        g1 = rand_element("xydplyplyplxydp")
        g2 = rand_element("xydplyplyplxydp")
        yield g1, g2

@pytest.mark.xsp2co1
def test_group(verbose = 0):
    unit = Xsp2_Co1()
    for ntest, (g1, g2) in enumerate(create_test_elements()):
        gm1 = Xsp2_Co1(g1)    
        gm2 = Xsp2_Co1(g2)    
        gm3_ref = Xsp2_Co1(g1 + g2)    
        gm3 = gm1 * gm2
        ok = gm3 == gm3_ref 
        if verbose or not ok:
            print("\nTEST %s" % (ntest+1))
            print("element g1 =", g1)
            print("element g2 =", g2)
            print("element g1 * g2")
            print("expected:", gm3_ref)
            print("obtained:", gm3)
            if not ok:
                ERR = "Multiplication in group G_{x1} failed"
                raise ValueError(ERR)
        gq1 = gm1.qs
        gq2 = gm2.qs
        gq3 = gm3.qs  
        gq3_ref = gq1 @ gq2
        sign_q  = int(gq3 == -gq3_ref)
        assert sign_q  or  gq3_ref == gq3       

        gleech1 = gm1.leech_op
        gleech2 = gm2.leech_op
        gleech3 = gm3.leech_op  
        gleech3_ref = gleech1 @ gleech2 / 8
        sign_leech  = int((gleech3 == -gleech3_ref).all())
        assert  (gleech3 @ gleech3.T == 64 * np.eye(24)).all()       
        assert sign_leech  or  (gleech3_ref == gleech3).all()

                
        gm1i =  gm1**(-1) 
        ok = gm1 * gm1i == unit
        if verbose or not ok:
            print("g1\n",  gm1) 
            print("g1 ** -1 obtained\n",  gm1**(-1)) 
            if not ok:            
                ERR = "Inversion in group G_{x1} failed"
                raise ValueError(ERR)
            
            



"""Accessing vectors of the representation of the monster

We deal with the rational representation [Seysen20]  of the monster 
group MM, which is  based on the Conway's construction [Con85] of
the monster, modulo various small integers p = 2**n-1, 2 <= n <= 8.
Here the integer p is called the modulus. 

All representations R_p of MM are 196884-dimensional matrix 
representations over the integers modulo p. Here the matrices
operate on vectors which are represented as arrays of integers.
The matrix operations on these vectors are performed by C functions.

For vectors there is an external representation as an array of 
integers of type uint8_t and there is also an internal representation. 
The internal representation is fairly close to the description of the 
basis vectors of the representation in [Seysen20] , so that matrix 
operations can take advantage of the internal structure of that 
repesentation.

The purpose of this module is access to the internal representation 
of such a vector and conversion between the external and the 
internal representation.  

There are versions optimized for 32-bit and for 64-bit computers. 
This version is optimized for {INT_BITS} bits. It supports 
moduli p = {P_LIST}.


References
----------
see file mmgroup.pdf
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import collections
"""TODO: write module documentation

"""


import warnings
from numbers import Integral
import numpy as np


from mmgroup.dev.mm_basics import mm_basics   


from mmgroup.dev.mm_basics.mm_basics import MM_Basics, MM_Const
from mmgroup.generate_c import c_snippet, TableGenerator
from mmgroup.generate_c import format_item

from mmgroup.bitfunctions import bitlen, hibit

from mmgroup.dev.mm_basics.mm_doc import __doc__ as MM_DOC
from mmgroup.dev.mm_basics.mm_basics import __doc__ as MM_BASICS_DOC


INT_BITS =  MM_Basics.INT_BITS 




#2345678901234567890123456789012345678901234567890123456789012345678901

class MM_IO24:
    """Generate code to access vectors of R_p in internal representation

    This class provides code generation functions for fast data
    transfer between a small array in a vector in R_p and an
    array of bytes. 

    We use the code generation mechanism in class  
    make_c_tabels.TableGenerator. The generated functions take
    an input parameter 'log_field_bits' coresponding to
    LOG_FIELD_BITS. This indicates the a bit field containing an
    entry contains 2**log_field_bits.
    """
    directives = {}

    @staticmethod
    def iter_ext_ABC():
        """Yield triples (tag, i, j) for external rep for tags A, B, C

        Triples are given in the order of the external representation
        for tags A, B and C.
        tag = 0, 1, 2 means 'A', 'B', 'C', respectively
        """
        for i in range(24):
            yield 0, i, i
        for tag in range(3):
            for i in range(24):
                for j in range(i):
                     yield tag, i, j

    @classmethod
    def abc_table(cls):
        """Table for expanding entries for tags 'A', 'B', 'C'.

        Entry k0 of the external representation of the monster is
        mapped to location k1 in the internal representation with
        k1 = (Table[k0] & 0x7ff) + k0 - 24. Entry k0 is also copied 
        to location k1 - 31 * (Table[k0] >> 11) of the internal rep.

        See comments 'Internal representation of a vector' and
        'External representation of a vector' for internal and
        external representation of a vector.
        """
        a = np.zeros(24 + 3 * 276, dtype = np.int16)
        for k0, (tag, i, j) in enumerate(cls.iter_ext_ABC()):
            k1 = (24 * tag + i) * 32 + j
            k1_entry = k1 - k0 + 24
            d_t  = i - j
            assert 0 <= k1_entry <  2048
            a[k0] = k1_entry  + (d_t << 11)
        return a
               


    @classmethod
    def reduce_table(cls):
        """Masks for reducing the fields of an uint_mmv_t modulo p

        Table entries 2*i-2 and 2*i-1 refer to modulus p = 2**i-1 for 
        i = 2,...,8. In the standard case, when i is not a power of 
        two, entries 2*i-2 and  2*i-1 have the following values:

        Index  Table entry
        2*i-2: A mask containg the value 1 in each field. 
        2*i-1: A mask containg the value 2**i-1 in each field. 

        If i is a power of two then table entry 2*i-1 is mask containing 
        the  value 2**(i/2)-1 in each field.

        The reason for that special case is that there is no space for 
        a carry bit between two adjacent fields if i is a power of two.
        In that case we need a special trick for detecting the value 
        2**i-i in a field.        
        """ 
        tbl = []
        for p_bits in range(2, 9):
            field_bits = p_bits
            while (field_bits & (field_bits - 1)): 
                field_bits += 1 
            mask_1 = sum(1 << i for i in range(0, INT_BITS, field_bits))
            q = p_bits >> (field_bits == p_bits)
            mask_q = ((1 << q) - 1) * mask_1
            tbl += [mask_1, mask_q]
        return tbl      
                    
    def __init__(self):
        """For code generation with class make_c_tables.TableGenerator

        """
        self.tables = {
            "MM_DOC": MM_DOC,
            "MM_BASICS_DOC": MM_BASICS_DOC,
            "MM_AUX": sys.modules[__name__],
            "MM_AUX_IO24": self,
            "MM_AUX_TBL_ABC" : self.abc_table(), 
            "MM_AUX_TBL_REDUCE": self.reduce_table(),
        }
        








"""Modelling vectors of the representation of the monster

We deal with the rational representation [Seys19] of the monster 
group MM, which is  based on the Conway's construction [Conw85] of 
the monster, modulo various small integers p = 2**n-1, n <= 8.
Here the integer p is called the modulus. See [Conw85], [Seys19], 
for details. We also refert to the documentation of the C modules 
mm_aux.c and mat24_functions.c in this package.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import os
from numbers import Integral
from random import randint, sample
import numpy as np
import pytest

from mmgroup.tests.spaces.sparse_mm_space import SparseMmSpace
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV
from mmgroup.tests.groups.mgroup_n import MGroupNWord 

V = SparseMmV
G = MGroupNWord

#################################################################################
# Test functions
#################################################################################


def direct_testdata():
    """Yield test triples (v, g, v*g)

    with v a vector, g a group element and v*g the expected product of v and g. 
    """
    space = V(7)
    g = G("d", 1) * G("y", 1)
    v = -space("Y", 0, 0)
    yield v*g, g**(-1), v
    g = G("t", 2) 
    v = 2*space('C', 0, 1)
    yield v*g, g**2, v

@pytest.mark.space
def test_sparse_direct():
    for v, g, expected in direct_testdata():
        if v * g != expected:
            print("v =", v)
            print("g =", g)
            print("1/g =", g**(-1))
            #assert g * g**(-1) == g**(-1) * g == g**0
            print("v * g expected =", expected)
            print("v * g obtained =", v * g)
            w = v
            txt = "v"
            for g0 in g:
                w *= g0
                print("%s * %s = %s" % (txt, g0, v))
                txt = " "
            raise ValueError("Bad vector multiplication result")
      



def mmv_data(n_tests, p, v_tags, g_tags, n_comp = 1):
    Vp = V(p)
    def rand_group_word(g_tags):
        tags = sample(g_tags, randint(3,5)) 
        return G([('x','r') for x in tags]) 
    def rand_vector(v_tags, n_comp):
        tags = [sample(v_tags, 1)[0] for i in range(n_comp)]
        return Vp([(x,'r') for x in tags])       
    
    for i in range(n_tests):
        g1 = rand_group_word(g_tags)
        g2 = rand_group_word(g_tags)
        yield rand_vector(v_tags, n_comp), g1, g2

def mmv_testdata(p):
    Vp = V(p)
    def mu(tag, *data):
        return Vp(tag, *data)
    def na(tag, *data):
        return G(tag, *data) 
    #n_group =  mmv_space.group
    testdata = [
        
        (mu('T',0,1) , na('t', 1) , na('t', 1) ),
        (mu('B',0,1) , na('t', 1) , na('d', 1177)),
        (mu('B',0,1) , na('t', 1) , na('p', 2221)),
        (3*mu('B',0,1) + mu('A',1,1), na('t', 1)*na('d', 1177) , na('y', 17)),
        (3*mu('B',0,1) + mu('A',1,1), na('t', 1)*na('p', 2221) , na('y', 17)),
        (3*mu('A', 1, 1), na('t', 1), na('t',1)),
        (2*mu('A', 0, 1), na('t', 1), na('t',1)),
        (mu('Y', 0, 0), na("d", 1), na("y", 1)),
        (mu('Y', 0, 0), na("y", 1), na("d", 1)),
        (mu('X', 0x6d3, 23), na("d", 1), na("d",0x331)),
        (mu('X', 0x6d3, 23), na("p", 7851), na("d",0x331)),
        (mu('X', 0x6d3, 23), na("d", 1), na("p", 197652)),
        (mu('X', 0x6d3, 23), na("p", 7851), na("p", 197652)),
    ]
    for d in testdata:
         yield d
    v_tags, g_tags = "ABCTXYZ", "dpxytl"
    for d in mmv_data(100, p,  v_tags, g_tags, 3):
         yield d
    print(".")
    v_tags, g_tags = "ABCTXYZ", "dpxytl"
    for d in mmv_data(3, p,  v_tags, g_tags, 1000):
         yield d
    print(".")



def short_testdata(mmv_space):
    #v_tags, g_tags = "ABCTXYZ", "dpxytl"
    v_tags, g_tags = "ABCTXYZ", "l"    
    v_tags, g_tags = "YZ", "l"    
    yield from mmv_data(5, mmv_space, v_tags, g_tags, n_comp = 1)

def l_testdata(mmv_space):
    v = mmv_space(("Y",0x400,0))
    group =  mmv_space.group
    yield v, group(("l",1)), group()

TEST_PAR = [
   # (3, l_testdata),
   # (7, short_testdata),
    (7, mmv_testdata)
]

@pytest.mark.space
@pytest.mark.parametrize("p, mmv_testdata", TEST_PAR )
def test_sparse_space(p, mmv_testdata):
    mmv = V(p)

    v = mmv('X', 0x6d3, 23)
    w =  mmv('Y', 0x126, 20)
    v += mmv('X', 0x126, 20) / 4
    maxlen = 0
    print( -17 * v / 8 + (w >> 16))
    for i, (v, g1, g2) in enumerate(mmv_testdata(p)):
        al = (v * g1) * g2
        ar = v * (g1 * g2)
        maxlen = max(maxlen, len(al))
        if al != ar:
            print("\nTest computing v * g1 * g2")
            print("v = ", v)
            print("g1 = ", g1)
            print("g2 = ", g2)
            print("v*g1 = ", v*g1)
            print("g1*g2 = ", g1*g2)
            print("(v*g1)*g2 = ", al)
            print("v*(g1*g2) = ", ar)
            print("Diff = ", al - ar)
            raise ValueError("Error in computing v * g1 * g2")
        g3 = (g1 * g2)**(-1)
        v_back = al * g3
        if v != v_back:
            print("\nTest computing v * g1 * g2 * g3 with g1 * g2 * g3 = 1")
            print("v = ", v)
            print("g1 = ", g1)
            print("g2 = ", g2)
            print("g3 = ", g3)
            print("g1*g2*g3 = ", g1*g2*g3)
            print("v*g1*g2 = ", al)
            print("v*g1*g2*g3 = ", v_back)
            print("Diff = ", v_back - v)
            raise ValueError("Error in computing v * g1 * g2 * g3")
            
        if i < 3:
            print("v = ", v)
            print("g1 =", g1)
            print("g2 =", g2)
            print("v*g1*g2 = ", al)
            print("")
            
    print("Max vector length is", maxlen)
    
      


"""Test C function dealing with Leech lattice vectors mod 3 of type 4

In this script we test functions dealing with vectors of type 4 in the 
Leech lattice modulo 3. These functions are implemented in file 
gen_leech3.c and available in the extension mmgroup.generators.

We use the terminology defined in
the document *The C interface of the mmgroup project*, 
section *Description of the mmgroup.generators extension*.
"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint, choices #, shuffle, sample
from functools import reduce
from operator import __or__
from numbers import Integral
from collections import defaultdict
from multiprocessing import Pool, TimeoutError

import numpy as np
import scipy
import scipy.special
from scipy.stats import chisquare
import pytest

from mmgroup import MM0
from mmgroup.mat24 import MAT24_ORDER, ploop_theta
from mmgroup.generators import gen_leech3to2_type4
from mmgroup.generators import gen_leech3_op_vector_word
from mmgroup.generators import gen_leech3_op_vector_atom
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_op_atom






#*************************************************************************
#** Convert type-4 vector mod 3 to type-4 vector mod 2
#************************************************************************/

from mmgroup.mat24 import bw24 as mat24_bw24
from mmgroup.mat24 import syndrome as mat24_syndrome
from mmgroup.mat24 import vect_to_cocode as mat24_vect_to_cocode
from mmgroup.mat24 import vect_to_gcode as mat24_vect_to_gcode
from mmgroup.mat24 import ploop_theta as MAT24_THETA_TABLE

def short_3_reduce(x):
    """Reduce coordinates of vector in Leech lattice mod 3

    The function reduces all coordinates of the vector x
    modulo 3, so that each coordinate has value 0, 1 or 2. Vector
    x must be given Leech lattice mod 3 encoding.
    """
    a = (x & (x >> 24)) & 0xffffff;
    x ^=  a | (a << 24);
    return x  & 0xffffffffffff;

def parity12(x):
    """Return parity of a 12-bit vector encoded as an integer"""
    x ^= x >> 6; x ^= x >> 3;
    return (0x96 >> (x & 7)) & 1;

def parity24(x):
    """Return parity of a 24-bit vector encoded as an integer"""
    x ^= x >> 12; x ^= x >> 6; x ^= x >> 3;
    return (0x96 >> (x & 7)) & 1;


def py_gen_leech3to2_type4(x):
    """Map type-4 vector from Leech lattice mod 3 to Leech lattice mod 2

    Here parameter ``x`` is a type-4 vector in the Leech lattice mod 3 
    in Leech lattice mod 3 encoding. 

    The function returns a type-4 vector in the Leech lattice mod 2
    corresponding to ``x`` in Leech lattice encoding.

    The result is unique. The function returns 0 if ``x`` is not of
    type 4 in the Leech lattice mod 3.

    This is a reference implementation for the C function
    **gen_leech3to2_type4** in file **gen_llech3.c**.
    """
    # uint_fast32_t gcodev, cocodev, h, w, w1, x1, syn, t, omega, res;
    x = short_3_reduce(x);
    # Let h be the support of x, i.e. the bit vector of nonzero
    # coordinates of the vector x (modulo 3)
    h = ((x >> 24) | x) & 0xffffff;
    # Let w1 and w2 be the number of indices with coordinate 1 and 2
    w = mat24_bw24(h);
    # Compute ``gcode`` and ``cocode`` for vector x. Return 0 if we 
    # detect that is not of type 4. If ``omega`` is odd then ``gcode`` 
    # has to be corrected by a term ``Omega``. At the end of the
    # switch statemnt, ``gcode`` might not correspond to a Golay
    # code vector; this means that x is not of type 4. If the scalar 
    # product of the result and ``Omega`` is one then we add a 
    # multiple of ``Omega`` to make that scalar product even.
    if w == 22:
            # type (5**1, 3**2, 1**21)
            syn = mat24_syndrome(x & 0xffffff, 0);
            gcodev = (x ^ syn) & 0xffffff;
            t = h & syn;
            cocodev = t | (0xffffff & ~h);
            if ((t == 0) or (t & (t-1))): return 0;
            omega = 0;
            #break;              
    elif w == 19:
            # type (3**5, 1**19)
            w1 = mat24_bw24(x & 0xffffff);
            x1 = x if (w1 & 1) else (x >> 24)
            x1 &= 0xffffff;
            syn = mat24_syndrome(x1, 0);
            cocodev = ~h & 0xffffff;
            if (syn & h): syn = cocodev;            
            gcodev = (x1 ^ syn) & 0xffffff;
            omega = 0;
            #break;
    elif w == 16:
            # type (2**16, 0**8)
            w1 = mat24_bw24(x & 0xffffff);
            if (w1 & 1): return 0;
            gcodev = h;
            omega = w1 >> 1;
            cocodev = x & 0xfffffff;
            #break;
    elif w in (10, 13):
            # type (4**1, 2**12, 0**11)
            # type (4**2, 2**10, 0**14)
            syn = mat24_syndrome(h & 0xffffff, 0);
            if ((h & syn) != syn): return 0;                  
            gcodev = h ^ syn;
            cocodev = syn | (x & ~syn & 0xffffff);
            w1 = mat24_bw24(cocodev);
            if (w1 & 1) : return 0;
            omega = (w1 >> 1) + parity24(syn & x) + w;
            #break; 
    elif w == 7:
            # type (6**1, 2**7, 0**16)
            syn = mat24_syndrome(h & 0xffffff, 0);
            if (syn & (syn - 1)): return 0;
            gcodev = h ^ syn;
            cocodev = (x & 0xffffff);
            w1 = mat24_bw24(cocodev);
            cocodev |=  (0 - (w1 & 1)) & syn;
            omega = ((w1 + 1) >> 1) + 1;
            #break; 
    elif w ==4:
            gcodev = 0;
            cocodev = h;
            omega = parity24(x);
            #break;
    elif w ==1:
            gcodev = cocodev = 0;
            omega = 1;
            #break;    
    else:
            return 0;        
    gcodev = mat24_vect_to_gcode(gcodev);
    if (gcodev & 0xfffff000): return 0;
    cocodev = mat24_vect_to_cocode(cocodev);
    cocodev ^= MAT24_THETA_TABLE(gcodev & 0x7ff) & 0xfff;
    # correct ``gcodev`` by term ``Omega`` if omega is odd
    gcodev ^= (omega & 1) << 11; 
    res = (gcodev << 12) ^ cocodev;
    # Correct an odd result
    if (w >= 19 and parity12(res & (res >> 12))): res ^= 0x800000;
    return res;



#####################################################################
# Creating test vectors


def rand_xsp2co1_elem(n):
    r"""Return random element of group ``G_x0`` as list of tuples

    The function returns a kind of random element of the subgroup
    ``G_x0`` of the monster as a list of tuples. That list may be
    passed to the constructor of class |MM| or |XSp2Co1|.

    Parameter ``n`` determines the number of generators 
    :math:`\xi^e, e = \pm1` in that list.
    """
    s1, l = [('p','r'),('y','r')],  [('l','n')]
    return s1 + (l + s1) * n






def create_test_elements():
    """Yield elements of the group ``G_x0`` as test data

    This generator yields elements of the group ``G_x0`` as test
    data. Each element is returned as a list o tuples as in
    function ``rand_xsp2co1_elem``. 
    """
    group_data = [
      [('x', 0x1f24), ('d', 0xf75)],
      [('x', 0x124), ('d', 0x555)],
      [('d', 0x124)],
      [('d', 0x800)],
      [('p', 187654344)],
      [('d', 0xd79), ('p', 205334671)],
      [('p', 205334671), ('d', 0xd79)],
      [('d', 0xd79), ('x', 0x1123)],
      [('y', 0x1d79)],
      [('y', 0x586)],
      [('l', 1)],
      [('l', 2)],
    ]
    for g in group_data:
        yield  g
    for n, imax  in [(1,50),(2,1000), (3,1000)]:
        for i in range(1 * imax):
            yield rand_xsp2co1_elem(n)


#####################################################################
# Test mapping of type 4 Leech vectors modulo 3 to vectors modulo 2
#####################################################################


def mul_v3(v3, g):
    """Multiply vector in Leech lattice mod 3 by element of ``G_x0``

    Here parameter ``v3`` is an element of the Leech lattice mod 3 in
    Leech lattice mod 3 encoding. Parameter ``g`` is an element of 
    the group ``G_x0`` encodes as an instance of class |MM| or |MM0|.

    The function returns the product ``v3 * g`` in  Leech lattice 
    mod 3 encoding.
    """
    result = gen_leech3_op_vector_word(v3, g._data, g.length)
    assert result & 0xffff000000000000 == 0, hex(result)
    return result
        
def mul_v2(v2, g):
    """Multiply vector in Leech lattice mod 2 by element of ``G_x0``

    Here parameter ``v2`` is an element of the Leech lattice mod 2 in
    Leech lattice encoding. Parameter ``g`` is an element of 
    the group ``G_x0`` endodes as an instance of class |MM| or |MM0|.

    The function returns the product ``v * g`` in  Leech lattice 
    encoding.
    """
    result = gen_leech2_op_word(v2, g._data, g.length)
    assert result & 0xfe000000 == 0, hex(result)
    return result

def v3_to_v2(v3):
    """Map type-4 vector from Leech lattice mod 3 to Leech lattice mod 2

    The operation of this function is equivalent to function
    ``py_gen_leech3to2_type4``. But this function is a wrapper for
    the C function ``gen_leech3to2_type4`` in file ``gen_leech3.c``. 
    """
    result = gen_leech3to2_type4(v3)
    assert result != 0, (str_v3(v3), weight_v3(v3), hex(result))
    return result


d_v3 = {0:0, 1:1, 0x1000000:2, 0x1000001:0}
def str_v3(v3):
    """Convert vector ``v3`` in Leech lattice mod 3 to a string.

    Here parameter ``v3`` is an element of the Leech lattice mod 3 in
    Leech lattice mod 3 encoding.    
    """
    l = [str(d_v3[(v3 >> i) & 0x1000001]) for i in range(24)]
    return "v3<%s>" % "".join(l)
    
w_v3 = {0:0, 1:1, 0x1000000:1, 0x1000001:0}
def weight_v3(v3):
    """Return weight of vector ``v3`` in Leech lattice mod 3.

    Here parameter ``v3`` is an element of the Leech lattice mod 3 in
    Leech lattice mod 3 encoding.    
    """
    return sum([w_v3[(v3 >> i) & 0x1000001] for i in range(24)])
    
    

@pytest.mark.gen_xi
def test_type4(verbose = 0):
    r"""Test conversion of type-4 vectors 

    Let \Omega be the type-4 vector in the Leech lattice corresponding
    to the standard frame. Let O_2 and O_3 be the images \Omega in the
    Leech lattice mod 2 and mod 3, respectively.

    For a set of elements g of the group ``G_x_0`` we convert O_3 * g to 
    a vector v2 in the Leech lattice mod 2  with function ``v3_to_v2`` 
    and we check that the result is equal to  O_2 * g. We use function
    ``create_test_elements`` for generating the elements g. 

    For the first few test data we also check this conversion against
    the reference implementation ``py_gen_leech3to2_type4``.
    """
    weights = defaultdict(int)
    for ntest, data in enumerate(create_test_elements()):
        g = MM0(data) 
        # Let \Omega be the standard type-4 vector in the Leech lattice
        # Let v3_st = \Omega in the Leech lattice mod 3
        v3_st = 1 << randint(0,23) 
        # Let v2_st = \Omega in the Leech lattice mod 3
        v2_st = 0x800000  # The standard type-4 vector \Omega      
        if verbose:
             print("\nTEST %s" % (ntest+1))
             print("v3_start = " , str_v3(v3_st))
             print("g =", g)
        v3 = mul_v3(v3_st, g)
        w = weight_v3(v3)  
        weights[w] += 1        
        v2_ref = mul_v2(v2_st, g) & 0xffffff
        v2 = v3_to_v2(v3) 
        ok = v2 == v2_ref 
        if  weights[w] <= 20:
             assert  v2 == py_gen_leech3to2_type4(v3)        
        if verbose or not ok:
            if not verbose:
                print("\nTEST %s" % (ntest+1))
                print("v3_start = " , str_v3(v3st))
                print("g =", g)
            print("v3 = v3_st*g =",str_v3(v3))
            print("weight =",w)
            print("v2 obtained= ", hex(v2))
            print("v2 expected= ", hex(v2_ref))
            if not ok:
                ERR = "Error in opation mod 3"
                raise ValueError(ERR)
    print("weights =", dict(weights))
    assert set(weights.keys()) == set([1,4,7,10,13,16,19,22])
    
    
    
#####################################################################
# Chisquare test of random type 4 Leech vectors modulo 3
#####################################################################



def binom(n, k):
    """Return binommial coeffient n choose k"""
    return int(scipy.special.binom(n, k) + 0.1)

# The dictionary contains the number DATA_GEOMETRY[w] of type-4 
# vectors in the Leech lattice modulo 3 of weight w. 
# This table is obtained from :cite:`Iva99`, Lemma 4.4.1
DATA_GEOMETRY = {
   1: 48, 
   7: 759 * 8 * 2**7,
  22: binom(24,3) * 3 * 2**12,
   4: binom(24,4) * 2**4,
  10: 759 * binom(16,2) * 2**9,
  13: 2576 * 12 * 2**12,
  19: binom(24,5) * 2**12,
  16: 759 * 16 * 2**11,
}  
 
# Number of type-4 vectors in the Leech lattice
NUM_LEECH_TYPE4 = 398034000 
# .. must be equal to the sum of the values in DATA_GEOMETRY
assert sum(DATA_GEOMETRY.values()) ==  NUM_LEECH_TYPE4
# Inverse of the number of vectors in Leech lattice mod 3
I_NUMV3 = 3.0**(-24)


# Assemble a dictionary DICT_P for a chisquare test.
# That dictionary maps a bit weight w to itself, if that bit weight
# occurs as a weight of a type-4 vector in the Leech lattice mod 3 
# with sufficiently high probability. It maps w to 1 if w occurs as
# such a weight with low probability.
# Let P be a dictionary that maps y the probability that a random 
# vector v in the Leech lattice mod 3 has type 4 and that 
# DICT_P[weight(v)] is equal to y. Let P[0] be the probablity that
# such a random vector y is not of type 4.
DICT_P = defaultdict(int)
P = defaultdict(float)
for w, num in DATA_GEOMETRY.items():
    p = num * I_NUMV3
    assert 0 <= p < 1
    DICT_P[w] = 1 if p < 1.0/4000 else w
    P[DICT_P[w]] += p
P[0] = 1.0 - NUM_LEECH_TYPE4 * I_NUMV3
DICT_P[0] = 0

P_MIN = min([x for x in P.values() if x > 0])

RANDMOD3 = [0, 1, 0x1000000]
def rand_v3():
    """Return a random vector in the space GF(3)^{24}.

    This random vector is encoded in **Leech lattice mod 3 encoding**.
    """
    randomlist = choices(RANDMOD3, k=24)
    return sum((x << i for i, x in enumerate(randomlist)))
    
def rand_v3_dict(n):
    """Create n random vectors and group the according to their weight

    We create n random vectors in GF(3)^{24}. We return a dictionary
    ``d`` with the following entries:

    d[0] counts the vectors not of type 4 in the Leech lattice mod 3.
    d[w] counts the vectors v of type 4 such that  DICT_P[weight(v)]
    is equal to w.

    Then the value d[i] should be about P[i] / n.    
    """
    d = defaultdict(int)
    for i in range(n):
        v3 = rand_v3()
        v2 = gen_leech3to2_type4(v3) 
        if (v2 == 0):
            d[0] += 1
        else: 
            w = mat24_bw24((v3 | (v3 >> 24)) & 0xffffff)
            d[DICT_P[w]] += 1
    return d



def chisquare_v3(obtained_dict, expected_dict):
    """Perform chi square test based on dictionaries

    Dictionary ``obtained_dict`` should have the same keys as
    dictionary ``expected_dict``. The function tests the hypothesis
    that the distribution of ``obtained_dict`` is propritional to
    that of ``expected_dict`` with a chisquare test.

    The function return a pair ``(chisq, p)`` with ``chisq`` the
    chi-squared test statistic, and ``p`` the p-value of the test.
    """
    f_obt, f_exp = [],[]
    for w in obtained_dict:
        assert w in expected_dict and expected_dict[w] > 0
    factor = sum(obtained_dict.values())/sum(expected_dict.values())
    for w in expected_dict:
        if expected_dict[w] > 0:
            f_exp.append(expected_dict[w] * factor)
            f_obt.append(obtained_dict[w])
    assert min(f_exp) > 9
    chisq, p = chisquare(f_obt, f_exp = f_exp)
    return chisq, p




@pytest.mark.gen_xi
def test_chisq_type4(n = 50000, verbose = 1):
    """Test distribution of weights of type-t vectors

    The function creates a list of n random vectors in GF(3)^{24}
    and groups them with respect to the property of having type 4
    in the Leech lattice and with respect to the weight as 
    described in function ``rand_v3_dict``.

    A chisquare test fails if the p-value is less than 0.01.
    We perform at most 4 chisquare test and raise ValueError
    if all of them fail.
    """
    p_min = 0.01
    print("Check distribution of type-4 vectors mod 3") 
    for i in range(4):
        d = rand_v3_dict(n)  
        chisq, p =  chisquare_v3(d, P)
        if verbose or i or p < p_min:
            print("Chisq = %.3f, p = %.4f" % (chisq, p))
        if p >= p_min: return
    raise ValueError("Chisquare test failed") 




